#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""Evohome RF - Schema processor."""

import logging
from typing import Tuple
import voluptuous as vol

from .const import (
    ATTR_CONTROLLER,
    ATTR_DEVICES,
    ATTR_DHW_SENSOR,
    ATTR_DHW_VALVE_HTG,
    ATTR_DHW_VALVE,
    ATTR_HTG_CONTROL,
    ATTR_ORPHANS,
    ATTR_STORED_HW,
    ATTR_SYSTEM,
    ATTR_UFH_CONTROLLERS,
    ATTR_ZONE_TYPE,
    ATTR_ZONE_SENSOR,
    ATTR_ZONES,
    DEFAULT_MAX_ZONES,
    ZONE_TYPE_SLUGS,
    _dev_mode_,
    id_to_address as addr,
)

DEVICE_ID_REGEXP = r"^[0-9]{2}:[0-9]{6}$"
DEVICE_ID = vol.Match(DEVICE_ID_REGEXP)

DOMAIN_ID_REGEXP = r"^[0-9A-F]{2}$"
DOMAIN_ID = vol.Match(DOMAIN_ID_REGEXP)

ZONE_IDX_REGEXP = r"^0[0-9AB]$"  # TODO: what if > 12 zones? (e.g. hometronics)
ZONE_IDX = vol.Match(ZONE_IDX_REGEXP)

ZONE_SCHEMA = vol.Schema(
    {
        vol.Required(ZONE_IDX): vol.Any(
            None,
            {
                vol.Optional(ATTR_ZONE_TYPE, default=None): vol.Any(
                    None, vol.Any(*list(ZONE_TYPE_SLUGS))
                ),
                vol.Optional(ATTR_ZONE_SENSOR, default=None): vol.Any(None, DEVICE_ID),
                vol.Optional(ATTR_DEVICES, default=[]): vol.Any(None, [DEVICE_ID]),
            },
        )
    }
)
SER2NET_SCHEMA = vol.Schema(
    {vol.Required("enabled"): bool, vol.Optional("socket", default="0.0.0.0:5000"): str}
)

# Config flags
CONFIG = "config"
DISABLE_DISCOVERY = "disable_discovery"
DISABLE_SENDING = "disable_sending"
ENFORCE_ALLOWLIST = "enforce_allowlist"
ENFORCE_BLOCKLIST = "enforce_blocklist"
EVOFW_FLAG = "evofw_flag"
INPUT_FILE = "input_file"
MAX_ZONES = "max_zones"
PACKET_LOG = "packet_log"
REDUCE_PROCESSING = "reduce_processing"
SERIAL_PORT = "serial_port"
SER2NET_RELAY = "ser2net_relay"
USE_NAMES = "use_names"  # use friendly device names from allow_list
USE_SCHEMA = "use_schema"
ALLOW_LIST = "allowlist"
BLOCK_LIST = "blocklist"


DONT_CREATE_MESSAGES = 3
DONT_CREATE_ENTITIES = 2
DONT_UPDATE_ENTITIES = 1

CONFIG_SCHEMA = vol.Schema(
    {
        vol.Optional(SERIAL_PORT): vol.Any(None, str),
        vol.Optional(DISABLE_SENDING, default=False): vol.Any(None, bool),
        vol.Optional(DISABLE_DISCOVERY, default=False): vol.Any(None, bool),
        vol.Optional(ENFORCE_ALLOWLIST, default=False): vol.Any(None, bool),
        vol.Optional(ENFORCE_BLOCKLIST, default=True): vol.Any(None, bool),
        vol.Optional(EVOFW_FLAG, default=None): vol.Any(None, str),
        vol.Optional(MAX_ZONES, default=DEFAULT_MAX_ZONES): vol.Any(None, int),
        vol.Optional(PACKET_LOG, default=None): vol.Any(None, str),
        vol.Optional(REDUCE_PROCESSING, default=0): vol.Any(None, int),
        # vol.Optional(SER2NET_RELAY): SER2NET_SCHEMA,
        vol.Optional(USE_NAMES, default=True): vol.Any(None, bool),
        vol.Optional(USE_SCHEMA, default=True): vol.Any(None, bool),
    },
    extra=vol.ALLOW_EXTRA,
)
DHW_SCHEMA = vol.Schema(
    {
        vol.Optional(ATTR_DHW_SENSOR): vol.Any(None, vol.Match(r"^07:[0-9]{6}$")),
        vol.Optional(ATTR_DHW_VALVE): vol.Any(None, vol.Match(r"^13:[0-9]{6}$")),
        vol.Optional(ATTR_DHW_VALVE_HTG): vol.Any(None, vol.Match(r"^13:[0-9]{6}$")),
    }
)
UFH_SCHEMA = vol.Schema({})
ORPHAN_SCHEMA = vol.Schema(
    {vol.Optional(ATTR_ORPHANS): vol.Any(None, vol.All(DEVICE_ID))}
)
SYSTEM_SCHEMA = vol.Schema(
    {
        vol.Required(ATTR_CONTROLLER): vol.Match(r"^(01|23):[0-9]{6}$"),
        vol.Optional(ATTR_SYSTEM, default={}): vol.Schema(
            {
                vol.Optional(ATTR_HTG_CONTROL): vol.Any(
                    None, vol.Match(r"^(10|13):[0-9]{6}$")
                ),
                vol.Optional(ATTR_ORPHANS): vol.Any(None, vol.All([DEVICE_ID])),
            }
        ),
        vol.Optional(ATTR_STORED_HW): vol.Any(None, DHW_SCHEMA),
        vol.Optional(ATTR_ZONES): vol.Any(
            None, vol.All(ZONE_SCHEMA, vol.Length(min=1, max=DEFAULT_MAX_ZONES))
        ),
        vol.Optional(ATTR_UFH_CONTROLLERS): vol.Any(None, [UFH_SCHEMA]),
    },
    extra=vol.ALLOW_EXTRA,
)
# GLOBAL_SCHEMA = vol.Schema(
#     vol.Any(SYSTEM_SCHEMA, vol.Length(min=0)), ORPHAN_SCHEMA, extra=vol.ALLOW_EXTRA
# )
KNOWNS_SCHEMA = vol.Schema(
    {
        vol.Optional(DEVICE_ID): vol.Any(
            None, {vol.Optional("name", default=None): vol.Any(None, str)}
        )  # , extra=vol.ALLOW_EXTRA
    }
)
FILTER_SCHEMA = vol.Schema(
    {
        vol.Optional(ALLOW_LIST): vol.Any(None, vol.All(KNOWNS_SCHEMA)),
        vol.Optional(BLOCK_LIST): vol.Any(None, vol.All(KNOWNS_SCHEMA)),
    }
)
# SCHEMA = vol.Schema(
#     {
#         vol.Optional("configuration"): CONFIG_SCHEMA,
#         vol.Optional("global_schema"): GLOBAL_SCHEMA,
#         vol.Optional("known_devices"): KNOWNS_SCHEMA,
#     }
# )
MONITOR_SCHEMA = vol.Schema(
    {
        vol.Optional("probe_system"): vol.Any(None, str),
        vol.Optional("execute_cmd"): vol.Any(None, str),
        vol.Optional(EVOFW_FLAG): vol.Any(None, str),
        vol.Optional(PACKET_LOG): vol.Any(None, str),
    }
)
PARSE_SCHEMA = vol.Schema({})
CLI_SCHEMA = vol.Schema({})

DEV_MODE = _dev_mode_

_LOGGER = logging.getLogger(__name__)
if DEV_MODE:
    _LOGGER.setLevel(logging.DEBUG)


def load_config(serial_port, input_file, **kwargs) -> Tuple[dict, list, list]:
    """Process the schema, and the configuration and return True if it is valid."""

    config = CONFIG_SCHEMA(kwargs.get(CONFIG, {}))
    allows = {}
    blocks = {}

    if config[ENFORCE_ALLOWLIST]:
        allows = KNOWNS_SCHEMA(kwargs.get(ALLOW_LIST, {}))
        config[ENFORCE_BLOCKLIST] = False
        if allows:
            _LOGGER.debug("An allowlist has been created, len = %s", len(allows))
        else:
            _LOGGER.warning("An empty allowlist was configured, so will be ignored")
            config[ENFORCE_ALLOWLIST] = False

    elif config[ENFORCE_BLOCKLIST]:
        blocks = KNOWNS_SCHEMA(kwargs.get(BLOCK_LIST, {}))
        if blocks:
            _LOGGER.debug("A blocklist has been created, len = %s", len(blocks))
        else:
            _LOGGER.debug("An empty blocklist was configured, so will be ignored")
            config[ENFORCE_BLOCKLIST] = False

    if serial_port and input_file:
        _LOGGER.warning(
            "Serial port was specified (%s), so input file (%s) will be ignored",
            serial_port,
            input_file,
        )
    elif serial_port is None:
        config[DISABLE_SENDING] = True

    if config[DISABLE_SENDING]:
        config[DISABLE_DISCOVERY] = True

    # if not kwargs.get(ALLOW_LIST, {}):
    #     config[USE_NAMES] = False

    return (config, allows, blocks)


def load_schema(gwy, **kwargs) -> Tuple[dict, dict]:
    """Process the schema, and the configuration and return True if it is valid."""
    # TODO: check a sensor is not a device in another zone

    schema = SYSTEM_SCHEMA(kwargs.get("schema", {})) if kwargs.get("schema") else {}

    device_ids = schema.get(ATTR_ORPHANS, [])  # TODO: clean up
    if device_ids is not None:
        for device_id in device_ids:
            gwy._get_device(addr(device_id))

    if not schema.get(ATTR_CONTROLLER):
        return ({}, KNOWNS_SCHEMA(kwargs.get(ALLOW_LIST, {})))

    schema = SYSTEM_SCHEMA(schema)

    ctl_id = schema[ATTR_CONTROLLER]
    ctl = gwy._get_device(addr(ctl_id), ctl_addr=addr(ctl_id))

    htg_ctl_id = schema[ATTR_SYSTEM].get(ATTR_HTG_CONTROL)
    if htg_ctl_id:
        ctl._evo._set_htg_control(gwy._get_device(addr(htg_ctl_id), ctl_addr=ctl))

    for device_id in schema[ATTR_SYSTEM].get(ATTR_ORPHANS, []):
        gwy._get_device(addr(device_id), ctl_addr=ctl)

    dhw = schema.get(ATTR_STORED_HW)
    if dhw:
        ctl._evo._set_dhw(ctl._evo._get_zone("HW"))

        dhw_sensor_id = dhw.get(ATTR_DHW_SENSOR)
        if dhw_sensor_id:
            ctl._evo._set_dhw_sensor(gwy._get_device(addr(dhw_sensor_id), ctl_addr=ctl))

        dhw_valve_id = dhw.get(ATTR_DHW_VALVE)
        if dhw_valve_id:
            ctl._evo._set_dhw_valve(gwy._get_device(addr(dhw_valve_id), ctl_addr=ctl))

        htg_valve_id = dhw.get(ATTR_DHW_VALVE_HTG)
        if htg_valve_id:
            ctl._evo._set_htg_valve(gwy._get_device(addr(htg_valve_id), ctl_addr=ctl))

    if ATTR_ZONES in schema:
        for zone_idx, attr in schema[ATTR_ZONES].items():
            zone = ctl._evo._get_zone(zone_idx, zone_type=attr.get(ATTR_ZONE_TYPE))

            sensor_id = attr.get(ATTR_ZONE_SENSOR)
            if sensor_id:
                zone._set_sensor(gwy._get_device(addr(sensor_id), ctl_addr=ctl))

            device_ids = attr.get(ATTR_DEVICES)  # TODO: clean up
            if device_ids is not None:
                for device_id in device_ids:
                    gwy._get_device(addr(device_id), ctl_addr=ctl, domain_id=zone_idx)

    # for ufh_ctl, ufh_schema in schema.get(ATTR_UFH_CONTROLLERS, []):
    #     dev = gwy._get_device(addr(ufh_ctl), ctl_addr=ctl)

    known_devices = KNOWNS_SCHEMA(kwargs.get(ALLOW_LIST, {}))
    known_devices.update(KNOWNS_SCHEMA(kwargs.get(BLOCK_LIST, {})))

    return (schema, known_devices)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""Evohome RF - The evohome-compatible zones."""

from abc import ABCMeta, abstractmethod
import asyncio
import logging
from typing import Optional

from .command import Schedule
from .const import (
    ATTR_DEVICES,
    ATTR_DHW_SENSOR,
    ATTR_DHW_VALVE,
    ATTR_DHW_VALVE_HTG,
    ATTR_HEAT_DEMAND,
    ATTR_SETPOINT,
    ATTR_TEMP,
    ATTR_OPEN_WINDOW,
    ATTR_ZONE_SENSOR,
    ATTR_ZONE_TYPE,
    # CODE_000C_DEVICE_TYPE,
    DEVICE_HAS_ZONE_SENSOR,
    DHW_STATE_MAP,
    DISCOVER_SCHEMA,
    DISCOVER_PARAMS,
    DISCOVER_STATUS,
    DISCOVER_ALL,
    ZONE_CLASS_MAP,
    ZONE_TYPE_MAP,
    ZONE_TYPE_SLUGS,
    ZONE_MODE_LOOKUP,
    ZONE_MODE_MAP,
    _dev_mode_,
)
from .devices import Device, Entity, _payload
from .exceptions import CorruptStateError
from .helpers import dtm_to_hex

DEV_MODE = _dev_mode_

_LOGGER = logging.getLogger(__name__)
if DEV_MODE:
    _LOGGER.setLevel(logging.DEBUG)


def _temp(value) -> str:
    """Return a two's complement Temperature/Setpoint."""
    if value is None:
        return "7FFF"

    try:
        _value = float(value)
    except (ValueError, TypeError):
        raise ValueError(f"Invalid temperature: {value}")

    if _value < 0:
        raise ValueError(f"Invalid temperature: {value}")

    return f"{int(_value*100):04X}"


class ZoneBase(Entity, metaclass=ABCMeta):
    """The Zone/DHW base class."""

    def __init__(self, evo, zone_idx) -> None:
        _LOGGER.debug("Creating a Zone: %s_%s (%s)", evo.id, zone_idx, self.__class__)
        super().__init__(evo._gwy)

        self.id = f"{evo.id}_{zone_idx}"
        self.idx = zone_idx

        self._name = None
        self._zone_type = None

        self._evo = evo
        self._ctl = evo._ctl

    # def __repr__(self) -> str:
    #     return json.dumps(self.schema, indent=2)

    def __str__(self) -> str:
        return f"{self.id} ({self._zone_type})"

    @abstractmethod
    def _discover(self, discover_flag=DISCOVER_ALL) -> None:
        raise NotImplementedError

    def _handle_msg(self, msg) -> bool:
        # super()._handle_msg(msg)
        pass

        # else:
        #     assert False, f"Unknown packet ({msg.verb}/{msg.code}) for {self.id}"

    def _send_cmd(self, code, **kwargs) -> None:
        dest = kwargs.pop("dest_addr", self._ctl.id)
        payload = kwargs.pop("payload", f"{self.idx}00")
        super()._send_cmd(code, dest, payload, **kwargs)

    @property
    @abstractmethod
    def schema(self) -> dict:
        """Return the schema of the zone/DHW (e.g. sensor_id, zone_type)."""
        raise NotImplementedError

    @property
    @abstractmethod
    def params(self) -> dict:
        """Return the configuration of the zone/DHW (e.g. min_temp, overrun)."""
        raise NotImplementedError

    @property
    @abstractmethod
    def status(self) -> dict:
        """Return the current state of the zone/DHW (e.g. setpoint, temperature)."""
        raise NotImplementedError

    @property
    @abstractmethod
    def sensor(self) -> Device:
        """Return the temperature sensor of the zone/DHW."""
        raise NotImplementedError

    @abstractmethod
    def _set_sensor(self, value: Device) -> None:
        """Set the temperature sensor for the zone (or for the DHW)."""
        raise NotImplementedError

    @property
    @abstractmethod
    def name(self) -> Optional[str]:
        """Return the name of the zone/DHW."""
        raise NotImplementedError

    @property
    @abstractmethod
    def mode(self) -> dict:
        """Return the operating mode of the zone/DHW ({mode, setpoint, until})."""
        raise NotImplementedError

    @property
    @abstractmethod
    def setpoint(self) -> Optional[float]:
        """Return the target temperature of the zone/DHW."""
        raise NotImplementedError

    # @setpoint.setter
    # @abstractmethod
    # def setpoint(self, value: float) -> None:  # TODO: also mode, name attrs
    #     """Set the target temperature of the zone/DHW."""
    #     raise NotImplementedError

    @property
    @abstractmethod
    def temperature(self) -> Optional[float]:
        """Return the measured temperature of the zone/DHW.

        The controller, not the temperature sensor, is the source of this data.
        """
        raise NotImplementedError

    # @property
    # @abstractmethod
    # def type(self) -> str:
    #     """Return the type of the zone/DHW (e.g. electric_zone, stored_dhw)."""
    #     raise NotImplementedError

    # @abstractmethod
    # def ._set_zone_type(self, value: str) -> None:
    #     """Set the type of the zone/DHW (e.g. electric_zone, stored_dhw)."""
    #     raise NotImplementedError


class HeatDemand:  # 3150
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._heat_demand = None

    def _handle_msg(self, msg) -> bool:
        super()._handle_msg(msg)

        if msg.code == "3150" and msg.verb == " I":
            self._heat_demand = msg

    @property
    def heat_demand(self) -> Optional[float]:  # 3150
        return _payload(self._heat_demand, "heat_demand")

    @property
    def status(self) -> dict:
        return {
            **super().status,
            "heat_demand": self.heat_demand,
        }


class DhwZone(ZoneBase, HeatDemand):
    """The DHW class."""

    def __init__(self, ctl, sensor=None, dhw_valve=None, htg_valve=None) -> None:
        super().__init__(ctl, "HW")

        ctl._set_dhw(self)

        self._sensor = None
        # self._dhw_valve = None
        # self._htg_valve = None

        self.heating_type = "DHW"

        self._dhw_mode = None
        self._dhw_params = None
        self._relay_demand = None
        self._relay_failsafe = None
        self._temp = None

        if sensor:
            self._set_sensor(sensor)
        if dhw_valve:
            self._set_dhw_valve(dhw_valve)
        if htg_valve:
            self._set_htg_valve(htg_valve)

    def _discover(self, discover_flags=DISCOVER_ALL) -> None:
        # super()._discover(discover_flag=discover_flag)

        # if False and _dev_mode_ and self.idx == "FA":  # dev/test code
        #     self.async_set_override(state="On")

        if discover_flags & DISCOVER_SCHEMA:
            [  # 000C: find the DHW relay(s), if any, see: CODE_000C_DEVICE_TYPE
                self._send_cmd("000C", payload=dev_type)
                for dev_type in ("000D", "000E", "010E")  # for DHW sensor, relay(s)
            ]

        if discover_flags & DISCOVER_PARAMS:
            for code in ("10A0",):
                self._send_cmd(code, payload="00")  # payload="00" (or "0000"), not "FA"

        if discover_flags & DISCOVER_STATUS:
            for code in ("1260", "1F41"):
                self._send_cmd(code, payload="00")  # payload="00" or "0000", not "FA"

    def _handle_msg(self, msg) -> bool:
        super()._handle_msg(msg)

        if msg.code == "0008":
            self._relay_demand = msg

        elif msg.code == "0009":
            self._relay_failsafe = msg

        elif msg.code == "10A0":
            self._dhw_params = msg

        elif msg.code == "1260":
            self._temp = msg

        elif msg.code == "1F41":
            self._setpoint_status = msg

        # elif msg.code in ("1100", "3150", "3B00"):
        #     pass

    @property
    def sensor(self) -> Device:
        return self._sensor

    def _set_sensor(self, device: Device) -> None:  # self._sensor
        """Set the temp sensor for this DHW system (07: only)."""

        if self._sensor != device and self._sensor is not None:
            raise CorruptStateError(
                f"{ATTR_ZONE_SENSOR} shouldn't change: {self._sensor} to {device}"
            )

        if not isinstance(device, Device) or device.type != "07":
            raise TypeError(f"{ATTR_ZONE_SENSOR} can't be: {device}")

        if self._sensor is None:
            self._sensor = device
            device._set_parent(self, domain="FA")

    @property
    def hotwater_valve(self) -> Device:
        return self._evo._dhw_valve

    @property
    def heating_valve(self) -> Device:
        return self._evo._htg_valve

    @property
    def relay_demand(self) -> Optional[float]:  # 0008
        return _payload(self._relay_demand, "relay_demand")

    @property  # only seen with FC, but seems should pair with 0008?
    def relay_failsafe(self) -> Optional[float]:  # 0009
        return _payload(self._relay_failsafe, "relay_failsafe")

    @property
    def dhw_config(self) -> Optional[dict]:  # 10A0
        return _payload(self._dhw_params)

    @property
    def mode(self) -> Optional[dict]:  # 1F41
        return _payload(self._dhw_mode)

    @property
    def name(self) -> str:
        return "Stored HW"

    @property
    def setpoint(self) -> Optional[float]:  # 1F41
        return _payload(self._dhw_mode, "setpoint")

    @property
    def temperature(self) -> Optional[float]:  # 1260
        return _payload(self._temp, "temperature")

    @property
    def schema(self) -> dict:
        """Return the stored HW's schema."""

        return {
            ATTR_DHW_SENSOR: self.sensor.id if self.sensor else None,
            ATTR_DHW_VALVE: self.hotwater_valve.id if self.hotwater_valve else None,
            ATTR_DHW_VALVE_HTG: self.heating_valve.id if self.heating_valve else None,
        }

    @property  # setpoint, config, mode (not schedule)
    def params(self) -> dict:
        return {
            "dhw_params": self._dhw_params,
        }

    @property  # temp, open_windows
    def status(self) -> dict:
        return {
            "temperature": self._temp,
            "dhw_mode": self._dhw_mode,
        }

    async def cancel_override(self) -> bool:  # 1F41
        """Reset the DHW to follow its schedule."""
        raise NotImplementedError

    async def set_override(self, mode=None, state=None, until=None) -> bool:
        """Force the DHW on/off for a duration, or indefinitely.

        Use until = ? for 1hr boost (obligates on)
        Use until = ? for until next scheduled on/off
        Use until = None for indefinitely
        """
        # 053  I --- 01:145038 --:------ 01:145038 1F41 012 00 01 04 FFFFFF 1E061B0607E4
        # 048  I --- 01:145038 --:------ 01:145038 1F41 012 00 00 04 FFFFFF 1E061B0607E4

        # if mode is None and until is None:
        #     mode = "00" if setpoint is None else "02"  # Follow, Permanent
        # elif mode is None:  # and until is not None
        #     mode = "04"  # Temporary
        # elif isinstance(mode, int):
        #     mode = f"{mode:02X}"
        # elif not isinstance(mode, str):
        #     raise TypeError("Invalid zone mode")
        # elif mode in ZONE_MODE_LOOKUP:
        #     mode = ZONE_MODE_LOOKUP[mode]

        # if mode not in ZONE_MODE_MAP:
        #     raise ValueError("Unknown zone mode")

        # if state is None and until is None:
        #     state = "01"
        # elif state is None:  # and until is not None
        #     state = "01"
        # elif isinstance(state, int):
        #     mode = f"{mode:02X}"
        # elif isinstance(state, bool):
        #     mode = "01" if mode is True else "00"
        # elif not isinstance(mode, str):
        #     raise TypeError("Invalid DHW state")
        # elif state in DHW_STATE_LOOKUP:
        #     state = DHW_STATE_LOOKUP[mode]

        if state not in DHW_STATE_MAP:
            raise ValueError("Unknown DHW state")

        if until is None:
            payload = f"00{state}{mode}FFFFFF"
        else:  # required only by: 04, Temporary, ignored by others
            payload = f"00{state}{mode}FFFFFF{dtm_to_hex(until)}"

        self._send_cmd("1F41", verb=" W", payload=payload)
        return False

    async def reset_config(self) -> bool:  # 10A0
        """Reset the DHW parameters to their default values."""
        raise NotImplementedError

    async def set_config(self, setpoint, overrun=None, differential=None) -> bool:
        """Set the DHW parameters."""
        raise NotImplementedError


class ZoneSchedule:
    """Evohome zones have a schedule."""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._schedule = Schedule(self)

    def _discover(self, discover_flag=DISCOVER_ALL) -> None:

        if discover_flag & DISCOVER_STATUS:
            # asyncio.create_task(self.get_schedule())  # 0404
            pass

    def _handle_msg(self, msg) -> bool:
        super()._handle_msg(msg)

        if msg.code == "0404" and msg.verb == "RP":
            _LOGGER.debug("Zone(%s).update: Received RP/0404 (schedule)", self.id)

    async def get_schedule(self, force_refresh=None) -> Optional[dict]:
        schedule = await self._schedule.get_schedule(force_refresh=force_refresh)
        if schedule:
            return schedule["schedule"]

    async def set_schedule(self, schedule) -> None:
        schedule = {"zone_idx": self.idx, "schedule": schedule}
        await self._schedule.set_schedule(schedule)

    @property
    def status(self) -> dict:
        return {
            **super().status,
            "schedule": self._schedule.schedule.get("schedule"),
        }


class Zone(ZoneSchedule, ZoneBase):
    """The Zone class."""

    def __init__(self, ctl, zone_idx, sensor=None, actuators=None) -> None:
        """Create a zone.

        The type of zone may not be known at instantiation. Even when it is known, zones
        are still created without a type before they are subsequently promoted, so that
        both schemes (e.g. eavesdropping, vs probing) are the same.

        In addition, an electric zone may subsequently turn out to be a zone valve zone.
        """
        # _LOGGER.debug("Creating a Zone: %s (%s)", zone_idx, self.__class__)
        super().__init__(ctl, zone_idx)

        # self.id = f"{ctl.id}_{zone_idx}"
        ctl.zones.append(self)
        ctl.zone_by_idx[zone_idx] = self
        # ctl.zone_by_name[self.name] = self

        self.devices = []
        self.device_by_id = {}
        self._sensor = None

        self._mode = None
        # self._name = None  # from super()
        self._setpoint = None
        self._temp = None
        self._zone_config = None

        # these needed here, as we can't use __init__
        self._mix_config = None
        self._ufh_setpoint = None
        self._window_open = None

        self._schedule = Schedule(self)

        if sensor:
            self._set_sensor(sensor)

    def _discover(self, discover_flag=DISCOVER_ALL) -> None:
        super()._discover(discover_flag=discover_flag)

        if DEV_MODE and self.idx == "99":  # dev/test code
            asyncio.create_task(  # TODO: test/dev only
                self.async_cancel_override()
                # self.async_set_override(
                #     setpoint=15.9,
                #     mode="AdvancedOverride",
                #     # until=dt_now() + td(minutes=120)
                # )
            )

        # TODO: add code to determine zone type if it doesn't have one, using 0005s

        [  # 000C: find the sensor and the actuators, if any
            self._send_cmd("000C", payload=f"{self.idx}{dev_type}")
            for dev_type in ("00", "04")  # CODE_000C_ZONE_TYPE
        ]  # TODO: use 08, not 00

        # start collecting the schedule
        # self._schedule.req_schedule()  # , restart=True) start collecting schedule

        for code in ("0004", "000A", "2349", "30C9"):  # sadly, no 3150
            self._send_cmd(code)  # , payload=self.idx)

    def _handle_msg(self, msg) -> bool:
        super()._handle_msg(msg)

        if isinstance(msg.payload, list):
            assert self.idx in [d["zone_idx"] for d in msg.payload]

        if msg.code == "0004":
            self._name = msg

        # not UFH (it seems), but ELE or VAL; and possibly a MIX support 0008 too
        elif msg.code in ("0008", "0009"):  # TODO: how to determine is/isn't MIX?
            assert msg.src.type in ("01", "13"), msg.src.type  # 01 as a stat
            assert self._zone_type in (None, "ELE", "VAL"), self._zone_type

            if self._zone_type is None:
                self._set_zone_type("ELE")  # might eventually be: "VAL"

        elif msg.code == "000A":
            self._zone_config = msg

        elif msg.code == "2309" and msg.verb in (" I", "RP"):  # setpoint
            assert msg.src.type == "01", "coding error"
            self._setpoint = msg

        elif msg.code == "2349" and msg.verb in (" I", "RP"):  # mode, setpoint
            assert msg.src.type == "01", "coding error"
            self._mode = msg
            self._setpoint = msg

        elif msg.code == "30C9" and msg.verb in (" I", "RP"):  # used by sensor matching
            assert msg.src.type in DEVICE_HAS_ZONE_SENSOR + ("01",), "coding error"
            self._temp = msg

        elif msg.code == "3150":  # TODO: and msg.verb in (" I", "RP")?
            assert msg.src.type in ("00", "02", "04", "13")
            assert self._zone_type in (None, "RAD", "UFH", "VAL")  # MIX/ELE don't 3150

            if msg.src.type in ("00", "02", "04", "13"):
                zone_type = ZONE_CLASS_MAP[msg.src.type]
                self._set_zone_type("VAL" if zone_type == "ELE" else zone_type)

        # elif "zone_idx" in msg.payload:
        #     pass

    @property
    def sensor(self) -> Device:
        return self._sensor

    def _set_sensor(self, device: Device):  # self._sensor
        """Set the temp sensor for this zone (one of: 01:, 03:, 04:, 12:, 22:, 34:)."""

        if self._sensor != device and self._sensor is not None:
            raise CorruptStateError(
                f"{ATTR_ZONE_SENSOR} shouldn't change: {self._sensor} to {device}"
            )

        sensor_types = ("00", "01", "03", "04", "12", "22", "34")
        if not isinstance(device, Device) or device.type not in sensor_types:
            raise TypeError(f"{ATTR_ZONE_SENSOR} can't be: {device}")

        if self._sensor is None:
            self._sensor = device
            device._set_parent(self)

    @property
    def heating_type(self) -> Optional[str]:
        """TODO.

        There are three ways to determine the type of a zone:
        1. Use a 0005 packet (deterministic)
        2. Eavesdrop (non-deterministic, slow to converge)
        3. via a config file (a schema)
        """

        if self._zone_type is not None:  # isinstance(self, ???)
            return ZONE_TYPE_MAP.get(self._zone_type)

        # TODO: actuators
        dev_types = [d.type for d in self.devices if d.type in ("00", "02", "04", "13")]

        if "02" in dev_types:
            zone_type = "UFH"
        elif "13" in dev_types:
            zone_type = "VAL" if "3150" in self._msgs else "ELE"
        # elif "??" in dev_types:  # TODO:
        #     zone_type = "MIX"
        elif "04" in dev_types or "00" in dev_types:
            # beware edge case: TRV as sensor for a non-RAD zone
            zone_type = "RAD"
        else:
            zone_type = None

        if zone_type is not None:
            self._set_zone_type(zone_type)

        return ZONE_TYPE_MAP.get(self._zone_type)

    def _set_zone_type(self, zone_type: str):  # self._zone_type
        """Set the zone's type, after validating it.

        There are two possible sources for the type of a zone:
        1. eavesdropping packet codes
        2. analyzing child devices

        Both will execute a zone.type = type (i.e. via this setter).
        """

        _type = ZONE_TYPE_SLUGS.get(zone_type, zone_type)
        if _type not in ZONE_CLASSES:
            raise ValueError(f"Not a known zone type: {zone_type}")

        if self._zone_type is not None:
            if self._zone_type != _type and (
                self._zone_type != "ELE" and _type != "VAL"
            ):
                raise CorruptStateError(
                    f"Zone {self} has a mismatched type: "
                    f"old={self._zone_type}, new={_type}"
                )

        self._zone_type = _type
        self.__class__ = ZONE_CLASSES[_type]
        _LOGGER.debug("Zone %s: type now set to %s", self.id, self._zone_type)

        self._discover()  # TODO: needs tidyup (ref #67)

    @property
    def heat_demand(self) -> Optional[float]:
        demands = [
            d.heat_demand
            for d in self.devices
            if hasattr(d, "heat_demand") and d.heat_demand is not None
        ]
        # return max(demands) if demands else None
        return round(sum(demands) / len(demands), 1) if demands else None

    @property
    def mode(self) -> Optional[dict]:  # 2349
        if not self._mode or self._mode.is_expired:
            return

        elif isinstance(self._mode.payload, dict):
            return _payload(self._mode)

        elif isinstance(self._mode.payload, list):
            tmp = [z for z in self._mode.payload if z["zone_idx"] == self.idx]
            return {k: v for k, v in tmp[0].items() if k[:1] != "_" and k != "zone_idx"}

    # async def get_name(self, force_refresh=None) -> Optional[str]:
    #     """Return the name of the zone."""
    #     if not force_refresh and self._name is not None:
    #         return _payload(self._name, "name")

    #     self._name = None
    #     self._send_cmd("0004", payload=f"{self.idx}00")
    #     while self._name is None:
    #         await asyncio.sleep(0.05)

    #     return _payload(self._name, "name")

    @property
    def name(self) -> Optional[str]:
        """Return the name of the zone."""
        return _payload(self._name, "name")

    # @name.setter
    # def name(self, value) -> Optional[str]:
    #     """Set the name of the zone."""
    #     return

    @property
    def setpoint(self) -> Optional[float]:  # 2309 (2349 is a superset of 2309)
        if not self._setpoint or self._setpoint.is_expired:
            return

        elif isinstance(self._setpoint.payload, dict):
            return _payload(self._setpoint, "setpoint")

        elif isinstance(self._setpoint.payload, list):
            _zone = [z for z in self._setpoint.payload if z["zone_idx"] == self.idx]
            return _zone[0]["setpoint"]

    @setpoint.setter
    def setpoint(self, value) -> None:
        """Set the target temperature, until the next scheduled setpoint."""
        if value is None:
            self.cancel_override()
        else:
            self.set_override(mode="advanced_override", setpoint=value)

    @property
    def temperature(self) -> Optional[float]:  # 30C9
        if not self._temp or self._temp.is_expired:
            if self.sensor and self.sensor.temperature:
                self._temp = self.sensor._temp

        if not self._temp or self._temp.is_expired:
            return

        elif isinstance(self._temp.payload, dict):
            return _payload(self._temp, "temperature")

        elif isinstance(self._temp.payload, list):
            tmp = [z for z in self._temp.payload if z["zone_idx"] == self.idx]
            return tmp[0]["temperature"]

    @property
    def zone_config(self) -> Optional[dict]:  # 000A
        if not self._zone_config or self._zone_config.is_expired:
            return

        elif isinstance(self._zone_config.payload, dict):
            return _payload(self._zone_config)

        elif isinstance(self._zone_config.payload, list):
            tmp = [z for z in self._zone_config.payload if z["zone_idx"] == self.idx]
            return {k: v for k, v in tmp[0].items() if k[:1] != "_" and k != "zone_idx"}

    def cancel_override(self) -> None:  # 2349
        """Revert to following the schedule."""
        self.set_override()

    def frost_protect(self) -> None:  # 2349
        """Set the zone to the lowest possible setpoint, indefinitely."""
        self.set_override(mode="permanent_override", setpoint=5)  # TODO

    def set_override(self, mode=None, setpoint=None, until=None) -> None:
        """Override the setpoint for a specified duration, or indefinitely.

        The setpoint has a resolution of 0.1 C. If a setpoint temperature is required,
        but none is provided, the controller will use the maximum possible value.

        The until has a resolution of 1 min.

        Incompatible combinations:
          - mode == Follow & setpoint not None (will silently ignore setpoint)
          - mode == Temporary & until is None (will silently ignore)
        """

        if mode is not None:
            if isinstance(mode, int):
                mode = f"{mode:02X}"
            elif not isinstance(mode, str):
                raise TypeError(f"Invalid zone mode: {mode}")
            if mode in ZONE_MODE_MAP:
                mode = ZONE_MODE_MAP["mode"]
            elif mode not in ZONE_MODE_LOOKUP:
                raise TypeError(f"Unknown zone mode: {mode}")
        elif until is None:  # mode is None
            mode = "advanced_override" if setpoint else "follow_schedule"
        else:  # if until is not None:
            mode = "temporary_override" if setpoint else "advanced_override"

        setpoint = _temp(setpoint)  # None means max, if a temp is required

        if until is None:
            mode = "advanced_override" if mode == "temporary_override" else mode

        mode = ZONE_MODE_LOOKUP[mode]

        if until is None:
            payload = f"{self.idx}{setpoint}{mode}FFFFFF"
        else:  # required only by temporary_override, ignored by others
            payload = f"{self.idx}{setpoint}{mode}FFFFFF{dtm_to_hex(until)}"

        self._send_cmd("2349", verb=" W", payload=payload)

    @property  # id, type
    def schema(self) -> dict:
        """Return the zone's schema."""

        return {
            ATTR_ZONE_TYPE: self.heating_type,
            ATTR_ZONE_SENSOR: self._sensor.id if self._sensor else None,
            ATTR_DEVICES: [d.id for d in self.devices],
        }

    @property  # setpoint, config, mode (not schedule)
    def params(self) -> dict:
        """Return the zone's configuration (excl. schedule)."""

        ATTR_NAME = "name"  # TODO
        ATTR_MODE = "mode"
        ATTR_CONFIG = "zone_config"

        return {
            ATTR_NAME: self.name,
            ATTR_MODE: self.mode,
            ATTR_CONFIG: self.zone_config,
        }

    @property
    def status(self) -> dict:
        """Return the zone's current state."""

        return {
            ATTR_SETPOINT: self.setpoint,
            ATTR_TEMP: self.temperature,
        }


class ZoneDemand:  # not all zone types call for heat
    """Not all zones call for heat."""

    def _discover(self, discover_flag=DISCOVER_ALL) -> None:
        super()._discover(discover_flag=discover_flag)

        if discover_flag & DISCOVER_STATUS:
            self._send_cmd("12B0")  # , payload=self.idx

    @property
    def heat_demand(self) -> Optional[float]:  # 3150
        demands = [
            d.heat_demand
            for d in self.devices
            if hasattr(d, "heat_demand") and d.heat_demand is not None
        ]
        return max(demands + [0]) if demands else None

    @property
    def status(self) -> dict:
        return {
            **super().status,
            ATTR_HEAT_DEMAND: self.heat_demand,
        }


class EleZone(Zone):  # Electric zones (do *not* call for heat)
    """For a small electric load controlled by a relay (never calls for heat)."""

    # def __init__(self, *args, **kwargs) -> None:  # can't use this here

    def _discover(self, discover_flag=DISCOVER_ALL) -> None:
        # super()._discover(discover_flag=discover_flag)
        self._send_cmd("000C", payload=f"{self.idx}11")

    def _handle_msg(self, msg) -> bool:
        super()._handle_msg(msg)

        if msg.code == "3150":  # ZV zones are Elec zones that also call for heat
            self._set_zone_type("VAL")

    @property
    def actuator_enabled(self) -> Optional[bool]:  # 3EF0
        return self._get_msg_value("3EF0")  # , "actuator_enabled"

    @property
    def actuator_state(self) -> Optional[float]:  # 3EF1
        return self._get_msg_value("3EF1")

    @property
    def status(self) -> dict:
        return {
            **super().status,
            "actuator_enabled": self.actuator_enabled,
            "actuator_state": self.actuator_state,
        }


class ValZone(ZoneDemand, EleZone):
    """For a motorised valve controlled by a BDR91 (will also call for heat)."""

    # def __init__(self, *args, **kwargs) -> None:  # can't use this here

    def _discover(self, discover_flag=DISCOVER_ALL) -> None:
        # super()._discover(discover_flag=discover_flag)
        self._send_cmd("000C", payload=f"{self.idx}0A")


class RadZone(ZoneDemand, Zone):
    """For radiators controlled by HR92s or HR80s (will also call for heat)."""

    # def __init__(self, *args, **kwargs) -> None:  # can't use this here

    def _discover(self, discover_flag=DISCOVER_ALL) -> None:
        # super()._discover(discover_flag=discover_flag)
        self._send_cmd("000C", payload=f"{self.idx}08")

    def _handle_msg(self, msg) -> bool:
        super()._handle_msg(msg)

        if msg.code == "12B0":
            self._window_open = msg

    @property
    def window_open(self) -> Optional[bool]:  # 12B0
        return _payload(self._window_open, "window_open")  # TODO: doesn't work >1 TRV

    @property
    def status(self) -> dict:
        return {
            **super().status,
            ATTR_OPEN_WINDOW: self.window_open,
        }


class UfhZone(ZoneDemand, Zone):
    """For underfloor heating controlled by an HCE80/HCC80 (will also call for heat)."""

    # def __init__(self, *args, **kwargs) -> None:  # can't use this here

    def _discover(self, discover_flag=DISCOVER_ALL) -> None:
        # super()._discover(discover_flag=discover_flag)
        self._send_cmd("000C", payload=f"{self.idx}09")

    def _handle_msg(self, msg) -> bool:
        super()._handle_msg(msg)

        if msg.code == "22C9" and msg.verb == " I":
            self._ufh_setpoint = msg

    @property
    def ufh_setpoint(self) -> Optional[float]:  # 22C9
        return _payload(self._ufh_setpoint, "setpoint")

    @property
    def status(self) -> dict:
        return {
            **super().status,
            "ufh_setpoint": self.ufh_setpoint,
        }


class MixZone(Zone):
    """For a modulating valve controlled by a HM80 (will also call for heat)."""

    # def __init__(self, *args, **kwargs) -> None:  # can't use this here

    def _discover(self, discover_flag=DISCOVER_ALL) -> None:
        # super()._discover(discover_flag=discover_flag)
        self._send_cmd("000C", payload=f"{self.idx}0B")

    def _handle_msg(self, msg) -> bool:
        super()._handle_msg(msg)

        if msg.code == "1030" and msg.verb == " I":
            self._mix_config = msg

    @property
    def mix_config(self) -> dict:
        return _payload(self._mix_config)

    @property
    def params(self) -> dict:
        return {**super().status, "mix_config": self.mix_config}


ZONE_CLASSES = {
    "RAD": RadZone,
    "ELE": EleZone,
    "VAL": ValZone,
    "UFH": UfhZone,
    "MIX": MixZone,
    "DHW": DhwZone,
}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""Evohome RF - Opentherm processor."""

import logging
import struct
from typing import Any

from .const import _dev_mode_

DEV_MODE = _dev_mode_

_LOGGER = logging.getLogger(__name__)
if DEV_MODE:
    _LOGGER.setLevel(logging.DEBUG)

# Data structure shamelessy copied, with thanks to @nlrb, from:
# github.com/nlrb/com.tclcode.otgw (ot_msg.js),

# Other code shamelessy copied, with thanks to @mvn23, from:
# github.com/mvn23/pyotgw (protocol.py),

OPENTHERM_MSG_TYPE = {
    0: "Read-Data",
    16: "Write-Data",
    32: "Invalid-Data",
    48: "-reserved-",
    64: "Read-Ack",
    80: "Write-Ack",
    96: "Data-Invalid",
    112: "Unknown-DataId",
}

OPENTHERM_MESSAGES = {
    # OpenTherm status flags [ID 0: Master status (HB) & Slave status (LB)]
    "status_flags": {
        "0x0100": {
            "en": "Central heating enable",
            "nl": "Centrale verwarming aan",
            "var": "StatusCHEnabled",
        },
        "0x0200": {"en": "DHW enable", "nl": "Tapwater aan", "var": "StatusDHWEnabled"},
        "0x0400": {
            "en": "Cooling enable",
            "nl": "Koeling aan",
            "var": "StatusCoolEnabled",
        },
        "0x0800": {
            "en": "Outside temp. comp. active",
            "nl": "Compenseren buitentemp.",
            "var": "StatusOTCActive",
        },
        "0x1000": {
            "en": "Central heating 2 enable",
            "nl": "Centrale verwarming 2 aan",
            "var": "StatusCH2Enabled",
        },
        "0x2000": {
            "en": "Summer/winter mode",
            "nl": "Zomer/winter mode",
            "var": "StatusSummerWinter",
        },
        "0x4000": {
            "en": "DHW blocking",
            "nl": "Tapwater blokkade",
            "var": "StatusDHWBlocked",
        },
        "0x0001": {
            "en": "Fault indication",
            "nl": "Fout indicatie",
            "var": "StatusFault",
        },  # no fault/fault
        "0x0002": {
            "en": "Central heating mode",
            "nl": "Centrale verwarming mode",
            "var": "StatusCHMode",
        },  # not active/active
        "0x0004": {
            "en": "DHW mode",
            "nl": "Tapwater mode",
            "var": "StatusDHWMode",
        },  # not active/active
        "0x0008": {
            "en": "Flame status",
            "nl": "Vlam status",
            "var": "StatusFlame",
        },  # flame off/on
        "0x0010": {
            "en": "Cooling status",
            "nl": "Status koelen",
            "var": "StatusCooling",
        },  # not active/active
        "0x0020": {
            "en": "Central heating 2 mode",
            "nl": "Centrale verwarming 2 mode",
            "var": "StatusCH2Mode",
        },  # not active/active
        "0x0040": {
            "en": "Diagnostic indication",
            "nl": "Diagnose indicatie",
            "var": "StatusDiagnostic",
        },  # no diagnostics/diagnostics event
    },
    # OpenTherm Master configuration flags [ID 2: master config flags (HB)]
    "Master_config_flags": {"0x0100": {"en": "Smart Power", "var": "ConfigSmartPower"}},
    # OpenTherm Slave configuration flags [ID 3: slave config flags (HB)]
    "Slave_Config_flags": {
        "0x0100": {"en": "DHW present", "var": "ConfigDHWpresent"},
        "0x0200": {
            "en": "Control type (modulating on/off)",
            "var": "ConfigControlType",
        },
        "0x0400": {"en": "Cooling supported", "var": "ConfigCooling"},
        "0x0800": {"en": "DHW storage tank", "var": "ConfigDHW"},
        "0x1000": {
            "en": "Master low-off & pump control allowed",
            "var": "ConfigMasterPump",
        },
        "0x2000": {"en": "Central heating 2 present", "var": "ConfigCH2"},
    },
    # OpenTherm fault flags [ID 5: Application-specific fault flags (HB)]
    "fault_flags": {
        "0x0100": {
            "en": "Service request",
            "nl": "Onderhoudsvraag",
            "var": "FaultServiceRequest",
        },
        "0x0200": {
            "en": "Lockout-reset",
            "nl": "Geen reset op afstand",
            "var": "FaultLockoutReset",
        },
        "0x0400": {
            "en": "Low water pressure",
            "nl": "Waterdruk te laag",
            "var": "FaultLowWaterPressure",
        },
        "0x0800": {
            "en": "Gas/flame fault",
            "nl": "Gas/vlam fout",
            "var": "FaultGasFlame",
        },
        "0x1000": {
            "en": "Air pressure fault",
            "nl": "Luchtdruk fout",
            "var": "FaultAirPressure",
        },
        "0x2000": {
            "en": "Water over-temperature",
            "nl": "Water te heet",
            "var": "FaultOverTemperature",
        },
    },
    # OpenTherm remote flags [ID 6: Remote parameter flags (HB)]
    "Remote_flags": {
        "0x0100": {"en": "DHW setpoint enable", "var": "RemoteDHWEnabled"},
        "0x0200": {"en": "Max. CH setpoint enable", "var": "RemoteMaxCHEnabled"},
        "0x0001": {"en": "DHW setpoint read/write", "var": "RemoteDHWReadWrite"},
        "0x0002": {"en": "Max. CH setpoint read/write", "var": "RemoteMaxCHReadWrite"},
    },
    # OpenTherm messages
    "messages": {
        "0": {"en": "Status", "dir": "R-", "val": "flag8", "flags": "StatusFlags"},
        "1": {
            "en": "Control setpoint",
            "nl": "Ketel doeltemperatuur",
            "dir": "-W",
            "val": "f8.8",
            "var": "ControlSetpoint",
            "sensor": "temperature",
        },
        "2": {
            "en": "Master configuration",
            "dir": "-W",
            "val": {"hb": "flag8", "lb": "u8"},
            "flags": "MasterConfigFlags",
            "var": {"lb": "MasterMemberId"},
        },
        "3": {
            "en": "Slave configuration",
            "dir": "R-",
            "val": {"hb": "flag8", "lb": "u8"},
            "flags": "SlaveConfigFlags",
            "var": {"lb": "SlaveMemberId"},
        },
        "4": {"en": "Remote command", "dir": "-W", "val": "u8", "var": "RemoteCommand"},
        "5": {
            "en": "Fault flags & OEM fault code",
            "dir": "R-",
            "val": {"hb": "flag8", "lb": "u8"},
            "var": {"lb": "OEMFaultCode"},
            "flags": "FaultFlags",
        },
        "6": {
            "en": "Remote parameter flags",
            "dir": "R-",
            "val": "flag8",
            "flags": "RemoteFlags",
        },
        "7": {
            "en": "Cooling control signal",
            "dir": "-W",
            "val": "f8.8",
            "var": "CoolingControlSignal",
            "sensor": "percentage",
        },
        "8": {
            "en": "Control setpoint central heating 2",
            "dir": "-W",
            "val": "f8.8",
            "var": "CH2ControlSetpoint",
            "sensor": "temperature",
        },
        "9": {
            "en": "Remote override room setpoint",
            "nl": "Overschreven kamer doeltemperatuur",
            "dir": "R-",
            "val": "f8.8",
            "var": "RemoteOverrideRoomSetpoint",
            "sensor": "temperature",
        },
        "10": {
            "en": "Number of transparent slave parameters (TSP) supported by slave",
            "dir": "R-",
            "val": "u8",
            "var": {"hb": "TSPNumber"},
        },
        "11": {
            "en": "Index number/value of referred-to transparent slave parameter (TSP)",
            "dir": "RW",
            "val": "u8",
            "var": {"hb": "TSPIndex", "lb": "TSPValue"},
        },
        "12": {
            "en": "Size of fault history buffer (FHB) supported by slave",
            "dir": "R-",
            "val": "u8",
            "var": {"hb": "FHBSize"},
        },
        "13": {
            "en": "Index number/value of referred-to fault history buffer (FHB) entry",
            "dir": "R-",
            "val": "u8",
            "var": {"hb": "FHBIndex", "lb": "FHBValue"},
        },
        "14": {
            "en": "Max. relative modulation level",
            "nl": "Max. relatief modulatie-niveau",
            "dir": "-W",
            "val": "f8.8",
            "var": "MaxRelativeModulationLevel",
            "sensor": "percentage",
        },
        "15": {
            "en": "Max. boiler capacity (kW) and modulation level setting (%)",
            "dir": "R-",
            "val": "u8",
            "var": {"hb": "MaxBoilerCapacity", "lb": "MinModulationLevel"},
        },
        "16": {
            "en": "Room setpoint",
            "nl": "Kamer doeltemperatuur",
            "dir": "-W",
            "val": "f8.8",
            "var": "CurrentSetpoint",
            "sensor": "temperature",
        },
        "17": {
            "en": "Relative modulation level",
            "nl": "Relatief modulatie-niveau",
            "dir": "R-",
            "val": "f8.8",
            "var": "RelativeModulationLevel",
            "sensor": "percentage",
        },
        "18": {
            "en": "Central heating water pressure",
            "nl": "Keteldruk",
            "dir": "R-",
            "val": "f8.8",
            "var": "CHWaterPressure",
            "sensor": "pressure",
        },
        "19": {
            "en": "DHW flow rate (litres/minute)",
            "dir": "R-",
            "val": "f8.8",
            "var": "DHWFlowRate",
            "sensor": "flow",
        },
        "20": {"en": "Day of week & time of day", "dir": "RW", "var": "DayTime"},
        "21": {"en": "Date", "dir": "RW", "val": "u8", "var": "Date"},
        "22": {"en": "Year", "dir": "RW", "val": "u16", "var": "Year"},
        "23": {
            "en": "Room setpoint central heating 2",
            "dir": "-W",
            "val": "f8.8",
            "var": "CH2CurrentSetpoint",
            "sensor": "temperature",
        },
        "24": {
            "en": "Room temperature",
            "nl": "Kamertemperatuur",
            "dir": "-W",
            "val": "f8.8",
            "var": "CurrentTemperature",
            "sensor": "temperature",
        },
        "25": {
            "en": "Boiler water temperature",
            "nl": "Ketelwatertemperatuur",
            "dir": "R-",
            "val": "f8.8",
            "var": "BoilerWaterTemperature",
            "sensor": "temperature",
        },
        "26": {
            "en": "DHW temperature",
            "nl": "Tapwatertemperatuur",
            "dir": "R-",
            "val": "f8.8",
            "var": "DHWTemperature",
            "sensor": "temperature",
        },
        "27": {
            "en": "Outside temperature",
            "nl": "Buitentemperatuur",
            "dir": "R-",
            "val": "f8.8",
            "var": "OutsideTemperature",
            "sensor": "temperature",
        },
        "28": {
            "en": "Return water temperature",
            "nl": "Retourtemperatuur",
            "dir": "R-",
            "val": "f8.8",
            "var": "ReturnWaterTemperature",
            "sensor": "temperature",
        },
        "29": {
            "en": "Solar storage temperature",
            "dir": "R-",
            "val": "f8.8",
            "var": "SolarStorageTemperature",
            "sensor": "temperature",
        },
        "30": {
            "en": "Solar collector temperature",
            "dir": "R-",
            "val": "f8.8",
            "var": "SolarCollectorTemperature",
            "sensor": "temperature",
        },
        "31": {
            "en": "Flow temperature central heating 2",
            "dir": "R-",
            "val": "f8.8",
            "var": "CH2FlowTemperature",
            "sensor": "temperature",
        },
        "32": {
            "en": "DHW 2 temperature",
            "dir": "R-",
            "val": "f8.8",
            "var": "DHW2Temperature",
            "sensor": "temperature",
        },
        "33": {
            "en": "Boiler exhaust temperature",
            "dir": "R-",
            "val": "s16",
            "var": "BoilerExhaustTemperature",
            "sensor": "temperature",
        },
        "48": {
            "en": "DHW setpoint boundaries",
            "dir": "R-",
            "val": "s8",
            "var": "DHWBounadries",
            "sensor": "temperature",
        },
        "49": {
            "en": "Max. central heating setpoint boundaries",
            "dir": "R-",
            "val": "s8",
            "var": "CHBoundaries",
            "sensor": "temperature",
        },
        "50": {
            "en": "OTC heat curve ratio upper & lower bounds",
            "dir": "R-",
            "val": "s8",
            "var": "OTCBoundaries",
        },
        "56": {
            "en": "DHW setpoint",
            "nl": "Tapwater doeltemperatuur",
            "dir": "RW",
            "val": "f8.8",
            "var": "DHWSetpoint",
            "sensor": "temperature",
        },
        "57": {
            "en": "Max. central heating water setpoint",
            "nl": "Max. ketel doeltemperatuur",
            "dir": "RW",
            "val": "f8.8",
            "var": "MaxCHWaterSetpoint",
            "sensor": "temperature",
        },
        "58": {
            "en": "OTC heat curve ratio",
            "dir": "RW",
            "val": "f8.8",
            "var": "OTCHeatCurveRatio",
            "sensor": "temperature",
        },
        # OpenTherm 2.3 IDs (70-91) for ventilation/heat-recovery applications
        "70": {
            "en": "Status ventilation/heat-recovery",
            "dir": "R-",
            "val": "flag8",
            "var": "VHStatus",
        },
        "71": {
            "en": "Control setpoint ventilation/heat-recovery",
            "dir": "-W",
            "val": "u8",
            "var": {"hb": "VHControlSetpoint"},
        },
        "72": {
            "en": "Fault flags/code ventilation/heat-recovery",
            "dir": "R-",
            "val": {"hb": "flag", "lb": "u8"},
            "var": {"lb": "VHFaultCode"},
        },
        "73": {
            "en": "Diagnostic code ventilation/heat-recovery",
            "dir": "R-",
            "val": "u16",
            "var": "VHDiagnosticCode",
        },
        "74": {
            "en": "Config/memberID ventilation/heat-recovery",
            "dir": "R-",
            "val": {"hb": "flag", "lb": "u8"},
            "var": {"lb": "VHMemberId"},
        },
        "75": {
            "en": "OpenTherm version ventilation/heat-recovery",
            "dir": "R-",
            "val": "f8.8",
            "var": "VHOpenThermVersion",
        },
        "76": {
            "en": "Version & type ventilation/heat-recovery",
            "dir": "R-",
            "val": "u8",
            "var": {"hb": "VHProductType", "lb": "VHProductVersion"},
        },
        "77": {
            "en": "Relative ventilation",
            "dir": "R-",
            "val": "u8",
            "var": {"hb": "RelativeVentilation"},
        },
        "78": {
            "en": "Relative humidity",
            "nl": "Luchtvochtigheid",
            "dir": "RW",
            "val": "u8",
            "var": {"hb": "RelativeHumidity"},
            "sensor": "humidity",
        },
        "79": {
            "en": "CO2 level",
            "nl": "CO2 niveau",
            "dir": "RW",
            "val": "u16",
            "var": "CO2Level",
            "sensor": "co2",
        },
        "80": {
            "en": "Supply inlet temperature",
            "dir": "R-",
            "val": "f8.8",
            "var": "SupplyInletTemperature",
            "sensor": "temperature",
        },
        "81": {
            "en": "Supply outlet temperature",
            "dir": "R-",
            "val": "f8.8",
            "var": "SupplyOutletTemperature",
            "sensor": "temperature",
        },
        "82": {
            "en": "Exhaust inlet temperature",
            "dir": "R-",
            "val": "f8.8",
            "var": "ExhaustInletTemperature",
            "sensor": "temperature",
        },
        "83": {
            "en": "Exhaust outlet temperature",
            "dir": "R-",
            "val": "f8.8",
            "var": "ExhaustOutletTemperature",
            "sensor": "temperature",
        },
        "84": {
            "en": "Actual exhaust fan speed",
            "dir": "R-",
            "val": "u16",
            "var": "ExhaustFanSpeed",
        },
        "85": {
            "en": "Actual inlet fan speed",
            "dir": "R-",
            "val": "u16",
            "var": "InletFanSpeed",
        },
        "86": {
            "en": "Remote parameter settings ventilation/heat-recovery",
            "dir": "R-",
            "val": "flag8",
            "var": "VHRemoteParameter",
        },
        "87": {
            "en": "Nominal ventilation value",
            "dir": "RW",
            "val": "u8",
            "var": "NominalVentilation",
        },
        "88": {
            "en": "TSP number ventilation/heat-recovery",
            "dir": "R-",
            "val": "u8",
            "var": {"hb": "VHTSPSize"},
        },
        "89": {
            "en": "TSP entry ventilation/heat-recovery",
            "dir": "RW",
            "val": "u8",
            "var": {"hb": "VHTSPIndex", "lb": "VHTSPValue"},
        },
        "90": {
            "en": "Fault buffer size ventilation/heat-recovery",
            "dir": "R-",
            "val": "u8",
            "var": {"hb": "VHFHBSize"},
        },
        "91": {
            "en": "Fault buffer entry ventilation/heat-recovery",
            "dir": "R-",
            "val": "u8",
            "var": {"hb": "VHFHBIndex", "lb": "VHFHBValue"},
        },
        # OpenTherm 2.2 IDs
        "100": {
            "en": "Remote override function",
            "dir": "R-",
            "val": {"hb": "flag8", "lb": "u8"},
            "var": {"hb": "RemoteOverrideFunction"},
        },
        "115": {
            "en": "OEM diagnostic code",
            "dir": "R-",
            "val": "u16",
            "var": "OEMDiagnosticCode",
        },
        "116": {
            "en": "Number of starts burner",
            "dir": "RW",
            "val": "u16",
            "var": "StartsBurner",
            "sensor": "counter",
        },
        "117": {
            "en": "Number of starts central heating pump",
            "dir": "RW",
            "val": "u16",
            "var": "StartsCHPump",
            "sensor": "counter",
        },
        "118": {
            "en": "Number of starts DHW pump/valve",
            "dir": "RW",
            "val": "u16",
            "var": "StartsHDWPump",
            "sensor": "counter",
        },
        "119": {
            "en": "Number of starts burner during DHW mode",
            "dir": "RW",
            "val": "u16",
            "var": "StartsBurnerDHW",
            "sensor": "counter",
        },
        "120": {
            "en": "Number of hours burner is in operation (i.e. flame on)",
            "dir": "RW",
            "val": "u16",
            "var": "HoursBurner",
            "sensor": "counter",
        },
        "121": {
            "en": "Number of hours central heating pump has been running",
            "dir": "RW",
            "val": "u16",
            "var": "HoursCHPump",
            "sensor": "counter",
        },
        "122": {
            "en": "Number of hours DHW pump has been running/valve has been opened",
            "dir": "RW",
            "val": "u16",
            "var": "HoursDHWPump",
            "sensor": "counter",
        },
        "123": {
            "en": "Number of hours DHW burner is in operation during DHW mode",
            "dir": "RW",
            "val": "u16",
            "var": "HoursDHWBurner",
            "sensor": "counter",
        },
        "124": {
            "en": "Opentherm version Master",
            "dir": "-W",
            "val": "f8.8",
            "var": "MasterOpenThermVersion",
        },
        "125": {
            "en": "Opentherm version Slave",
            "dir": "R-",
            "val": "f8.8",
            "var": "SlaveOpenThermVersion",
        },
        "126": {
            "en": "Master product version and type",
            "dir": "-W",
            "val": "u8",
            "var": {"hb": "MasterProductType", "lb": "MasterProductVersion"},
        },
        "127": {
            "en": "Slave product version and type",
            "dir": "R-",
            "val": "u8",
            "var": {"hb": "SlaveProductType", "lb": "SlaveProductVersion"},
        },
        # ZX-DAVB extras
        "113": {
            "en": "Number of un-successful burner starts",
            "dir": "RW",
            "val": "u16",
            "var": "BadStartsBurner?",
            "sensor": "counter",
        },
        "114": {
            "en": "Number of times flame signal was too low",
            "dir": "RW",
            "val": "u16",
            "var": "LowSignalsFlame?",
            "sensor": "counter",
        },
    },
}


def parity(x: int) -> int:
    """Make this the docstring."""
    shiftamount = 1
    while x >> shiftamount:
        x ^= x >> shiftamount
        shiftamount <<= 1
    return x & 1


def ot_msg_value(val_seqx, val_type) -> Any:
    """Make this the docstring."""

    def _get_flag8(byte, *args) -> list:
        """Split a byte (as a str) into a list of 8 bits (1/0)."""
        ret = [0] * 8
        byte = bytes.fromhex(byte)[0]
        for i in range(0, 8):
            ret[i] = byte & 1
            byte = byte >> 1
        return ret

    def _get_u8(byte, *args) -> int:
        """Convert a byte (as a str) into an unsigned int."""
        return struct.unpack(">B", bytes.fromhex(byte))[0]

    def _get_s8(byte, *args) -> int:
        """Convert a byte (as a str) into a signed int."""
        return struct.unpack(">b", bytes.fromhex(byte))[0]

    def _get_f8_8(msb, lsb) -> float:
        """Convert 2 bytes (as strs) into an OpenTherm f8_8 (float) value."""
        return float(_get_s16(msb, lsb) / 256)

    def _get_u16(msb, lsb) -> int:
        """Convert 2 bytes (as strs) into an unsigned int."""
        buf = struct.pack(">BB", _get_u8(msb), _get_u8(lsb))
        return int(struct.unpack(">H", buf)[0])

    def _get_s16(msb, lsb) -> int:
        """Convert 2 bytes (as strs) into a signed int."""
        buf = struct.pack(">bB", _get_s8(msb), _get_u8(lsb))
        return int(struct.unpack(">h", buf)[0])

    DATA_TYPES = {
        "flag8": _get_flag8,
        "u8": _get_u8,
        "s8": _get_s8,
        "f8.8": _get_f8_8,
        "u16": _get_u16,
        "s16": _get_s16,
    }

    if val_type in DATA_TYPES:
        return DATA_TYPES[val_type](val_seqx[:2], val_seqx[2:])
    return val_seqx


# See: https://www.opentherm.eu/request-details/?post_ids=2944
#
# ID0:HB0: Master status: CH enable
# ID0:HB1: Master status: DHW enable
# ID0:HB2: Master status: Cooling enable
# ID0:HB3: Master status: OTC active
# ID0:HB5: Master status: Summer/winter mode
# ID0:HB6: Master status: DHW blocking

# ID0:LB0: Slave Status: Fault indication
# ID0:LB1: Slave Status: CH mode
# ID0:LB2: Slave Status: DHW mode
# ID0:LB3: Slave Status: Flame status

# ID1: Control Setpoint i.e. CH water temperature Setpoint (°C)

# ID2:HB0: Master configuration: Smart power
# ID2:LB: Master MemberID Code

# ID3:HB0: Slave configuration: DHW present
# ID3:HB1: Slave configuration: Control type
# ID3:HB4: Slave configuration: Master low-off&pump control

# ID5:HB0: Service request
# ID5:HB1: Lockout-reset
# ID5:HB2: Low water pressure
# ID5:HB3: Gas/flame fault
# ID5:HB4: Air pressure fault
# ID5:HB5: Water over-temperature
# ID5:LB: OEM fault code

# ID6:HB0: Remote boiler parameter transfer-enable: DHW setpoint
# ID6:HB1: Remote boiler parameter transfer-enable: max. CH setpoint
# ID6:LB0: Remote boiler parameter read/write: DHW setpoint
# ID6:LB1: Remote boiler parameter read/write: max. CH setpoint

# ID9: Remote override room Setpoint
# ID10: Number of Transparent-Slave-Parameters supported by slave
# ID12: Size of Fault-History-Buffer supported by slave
# ID14: Maximum relative modulation level setting (%)
# ID16: Room Setpoint (°C)
# ID17: Relative Modulation Level (%)
# ID18: Water pressure in CH circuit (bar)
# ID19: Water flow rate in DHW circuit. (litres/minute)
# ID24: Room temperature (°C)
# ID25: Boiler flow water temperature (°C)
# ID26: DHW temperature (°C)
# ID27: Outside temperature (°C)
# ID28: Return water temperature (°C)
# ID48: DHW Setpoint upper & lower bounds for adjustment (°C)
# ID49: Max CH water Setpoint upper & lower bounds for adjustment (°C)
# ID56: DHW Setpoint (°C) (Remote parameter 1)
# ID57: Max CH water Setpoint (°C) (Remote parameters 2)

# ID126: Master product version number and type
# ID127: Slave product version number and type



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""Evohome RF - discovery scripts."""

import asyncio
import json
import logging
from typing import Any, List

from .command import Command, Priority
from .const import _dev_mode_, CODE_SCHEMA, DEVICE_TABLE, Address
from .exceptions import ExpiredCallbackError
from .ramses import HINTS_CODE_SCHEMA as HINTS_CODES

EXECUTE_CMD = "execute_cmd"
GET_FAULTS = "get_faults"
GET_SCHED = "get_schedule"
SET_SCHED = "set_schedule"

SCAN_DISC = "scan_disc"
SCAN_FULL = "scan_full"
SCAN_HARD = "scan_hard"
SCAN_XXXX = "scan_xxxx"


DEV_MODE = _dev_mode_

_LOGGER = logging.getLogger(__name__)
if DEV_MODE:
    _LOGGER.setLevel(logging.DEBUG)


async def _cmd(gwy, *args, **kwargs) -> None:
    qos = kwargs.pop("qos", {})
    await gwy.msg_protocol.send_data(Command(*args, **kwargs, qos=qos))


async def spawn_execute_cmd(gwy, **kwargs):
    if kwargs.get(EXECUTE_CMD):  # e.g. "RQ 01:145038 1F09 00"
        cmd = kwargs[EXECUTE_CMD]

        qos = {"priority": Priority.HIGH, "retries": 10}
        try:
            cmd = Command(cmd[:2], cmd[3:12], cmd[13:17], cmd[18:], qos=qos)
        except ValueError as err:
            _LOGGER.warning(
                "Execute: Command is invalid, and will be ignored: '%s' (%s)", cmd, err
            )
        else:
            await gwy.msg_protocol.send_data(cmd)


async def spawn_monitor_scripts(gwy, **kwargs) -> List[Any]:
    tasks = []

    if kwargs.get(EXECUTE_CMD):
        await spawn_execute_cmd(gwy, **kwargs)  # TODO: wrap in a try?

    if kwargs.get("poll_devices"):
        tasks += [poll_device(gwy, d) for d in kwargs["poll_devices"]]

    gwy._tasks.extend(tasks)
    return tasks


async def spawn_execute_scripts(gwy, **kwargs) -> List[Any]:

    # this is to ensure the gateway interface has fully woken
    if not kwargs.get(EXECUTE_CMD) and gwy._include:
        dev_id = next(iter(gwy._include))
        qos = {"priority": Priority.HIGH, "retries": 5}
        await gwy.msg_protocol.send_data(Command("RQ", dev_id, "0016", "00FF", qos=qos))

    tasks = []

    if kwargs.get(EXECUTE_CMD):  # TODO: wrap in a try?
        await spawn_execute_cmd(gwy, **kwargs)

    if kwargs.get(GET_FAULTS):
        tasks += [asyncio.create_task(get_faults(gwy, kwargs[GET_FAULTS]))]

    if kwargs.get(GET_SCHED) and kwargs[GET_SCHED][0]:
        tasks += [asyncio.create_task(get_schedule(gwy, *kwargs[GET_SCHED]))]

    if kwargs.get(SET_SCHED) and kwargs[SET_SCHED][0]:
        tasks += [asyncio.create_task(set_schedule(gwy, *kwargs[SET_SCHED]))]

    if kwargs.get(SCAN_DISC):
        tasks += [asyncio.create_task(scan_disc(gwy, d)) for d in kwargs[SCAN_DISC]]

    if kwargs.get(SCAN_FULL):
        tasks += [asyncio.create_task(scan_full(gwy, d)) for d in kwargs[SCAN_FULL]]

    if kwargs.get(SCAN_HARD):
        tasks += [asyncio.create_task(scan_hard(gwy, d)) for d in kwargs[SCAN_HARD]]

    if kwargs.get(SCAN_XXXX):
        tasks += [asyncio.create_task(scan_xxxx(gwy, d)) for d in kwargs[SCAN_XXXX]]

    gwy._tasks.extend(tasks)
    return tasks


async def periodic(gwy, cmd, count=1, interval=None):
    async def _periodic():
        await asyncio.sleep(interval)
        await gwy.msg_protocol.send_data(cmd)

    if interval is None:
        interval = 0 if count == 1 else 60

    if count <= 0:
        while True:
            await _periodic()
    else:
        for _ in range(count):
            await _periodic()


async def schedule_task(delay, func, *args, **kwargs) -> Any:
    """Start a coro after delay seconds."""

    async def scheduled_func(delay, func, *args, **kwargs):
        await asyncio.sleep(delay)
        await func(*args, **kwargs)

    return asyncio.create_task(scheduled_func(delay, func, *args, **kwargs))


async def get_faults(gwy, ctl_id: str):
    ctl_addr = Address(id=ctl_id, type=ctl_id[:2])
    device = gwy._get_device(ctl_addr, ctl_addr=ctl_addr)

    try:
        await device._evo.get_fault_log()  # 0418
    except ExpiredCallbackError as exc:
        _LOGGER.error("get_faults(): Function timed out: %s", exc)


async def get_schedule(gwy, ctl_id: str, zone_idx: str) -> None:
    ctl_addr = Address(id=ctl_id, type=ctl_id[:2])
    zone = gwy._get_device(ctl_addr, ctl_addr=ctl_addr)._evo._get_zone(zone_idx)

    try:
        await zone.get_schedule()
    except ExpiredCallbackError as exc:
        _LOGGER.error("get_schedule(): Function timed out: %s", exc)


async def set_schedule(gwy, ctl_id, schedule) -> None:
    schedule = json.load(schedule)
    zone_idx = schedule["zone_idx"]

    ctl_addr = Address(id=ctl_id, type=ctl_id[:2])
    zone = gwy._get_device(ctl_addr, ctl_addr=ctl_addr)._evo._get_zone(zone_idx)

    try:
        await zone.set_schedule(schedule["schedule"])  # 0404
    except ExpiredCallbackError as exc:
        _LOGGER.error("set_schedule(): Function timed out: %s", exc)


def poll_device(gwy, dev_id) -> List[Any]:
    _LOGGER.warning("poll_device() invoked...")

    qos = {"priority": Priority.LOW, "retries": 0}
    if "poll_codes" in DEVICE_TABLE.get(dev_id[:2]):
        codes = DEVICE_TABLE[dev_id[:2]]["poll_codes"]
    else:
        codes = ["0016", "1FC9"]

    tasks = []

    for code in codes:
        cmd = Command("RQ", dev_id, code, "00", qos=qos)
        tasks.append(asyncio.create_task(periodic(gwy, cmd, count=0)))

        cmd = Command("RQ", dev_id, code, "0000", qos=qos)
        tasks.append(asyncio.create_task(periodic(gwy, cmd, count=0)))

    gwy._tasks.extend(tasks)
    return tasks


async def scan_disc(gwy, dev_id: str):
    _LOGGER.warning("scan_quick() invoked...")

    device = gwy._get_device(Address(id=dev_id, type=dev_id[:2]))  # not always a CTL
    device._discover()  # discover_flag=DISCOVER_ALL)


async def scan_full(gwy, dev_id: str):
    _LOGGER.warning("scan_full() invoked - expect a lot of Warnings")

    qos = {"priority": Priority.LOW, "retries": 0}
    for code in sorted(HINTS_CODES):
        if code == "0005":
            for zone_type in range(20):  # known up to 18
                await _cmd(gwy, "RQ", dev_id, code, f"00{zone_type:02X}", qos=qos)
            continue

        elif code == "000C":
            for zone_idx in range(16):  # also: FA-FF?
                await _cmd(gwy, "RQ", dev_id, code, f"{zone_idx:02X}00", qos=qos)
            continue

        if code == "0016":
            qos_alt = {"priority": Priority.LOW, "retries": 5}
            await _cmd(gwy, "RQ", dev_id, code, "0000", qos=qos_alt)
            continue

        elif code == "0404":
            await _cmd(gwy, "RQ", dev_id, code, "00200008000100", qos=qos)

        elif code == "0418":
            for log_idx in range(2):
                await _cmd(gwy, "RQ", dev_id, code, f"{log_idx:06X}", qos=qos)
            continue

        elif code == "1100":
            await _cmd(gwy, "RQ", dev_id, code, "FC", qos=qos)

        elif code == "2E04":
            await _cmd(gwy, "RQ", dev_id, code, "FF", qos=qos)

        elif code == "3220":
            for data_id in ("00", "03"):  # these are mandatory READ_DATA data_ids
                await _cmd(gwy, "RQ", dev_id, code, f"0000{data_id}0000", qos=qos)

        elif code in CODE_SCHEMA and CODE_SCHEMA[code].get("rq_len"):
            rq_len = CODE_SCHEMA[code].get("rq_len") * 2
            await _cmd(gwy, "RQ", dev_id, code, f"{0:0{rq_len}X}", qos=qos)

        else:
            await _cmd(gwy, "RQ", dev_id, code, "0000", qos=qos)

    # these are possible/difficult codes
    qos = {"priority": Priority.LOW, "retries": 3}
    for code in ("0150", "2389"):
        await _cmd(gwy, "RQ", dev_id, code, "0000", qos=qos)


async def scan_hard(gwy, dev_id: str):
    _LOGGER.warning("scan_deep() invoked - expect some Warnings")

    qos = {"priority": Priority.LOW, "retries": 0}
    for code in range(0x4000):
        await _cmd(gwy, "RQ", dev_id, f"{code:04X}", "0000", qos=qos)


async def scan_xxxx(gwy, dev_id: str):
    _LOGGER.warning("scan_xxx() invoked - expect a lot of nonsense")

    qos = {"priority": Priority.LOW, "retries": 3}
    for idx in range(0x10):
        await _cmd(gwy, " W", dev_id, "000E", f"{idx:02X}0050", qos=qos)
        await _cmd(gwy, "RQ", dev_id, "000E", f"{idx:02X}00C8", qos=qos)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""Evohome RF - a RAMSES-II protocol decoder & analyser.

Works with (amongst others):
- evohome (up to 12 zones)
- sundial (up to 2 zones)
- chronotherm (CM60xNG can do 4 zones)
- hometronics (16? zones)
"""

import asyncio
from collections import deque
import json
import logging
import os
import signal
from threading import Lock
from typing import Callable, Dict, List, Tuple  # Any, Tuple

from .const import _dev_mode_, ATTR_ORPHANS
from .devices import DEVICE_CLASSES, Device
from .message import process_msg
from .packet import _PKT_LOGGER as pkt_logger, set_pkt_logging
from .protocol import create_msg_stack
from .transport import POLLER_TASK, create_pkt_stack
from .schema import (
    load_config,
    load_schema,
    DISABLE_DISCOVERY,
    DONT_CREATE_MESSAGES,
    PACKET_LOG,
    REDUCE_PROCESSING,
    USE_NAMES,
)
from .systems import SYSTEM_CLASSES, System, SystemBase
from .version import __version__  # noqa: F401

DEV_MODE = _dev_mode_
VERSION = __version__

_LOGGER = logging.getLogger(__name__)
if DEV_MODE:
    _LOGGER.setLevel(logging.DEBUG)


class GracefulExit(SystemExit):
    code = 1


class Gateway:
    """The gateway class."""

    def __init__(self, serial_port, input_file=None, loop=None, **kwargs) -> None:
        """Initialise the class."""

        if kwargs.pop("debug_mode", None):
            _LOGGER.setLevel(logging.DEBUG)  # should be INFO?
        _LOGGER.debug("Starting evohome_rf, **kwargs = %s", kwargs)

        self._loop = loop if loop else asyncio.get_running_loop()
        self._tasks = []
        self._setup_event_handlers()

        self.serial_port = serial_port
        self._input_file = input_file

        (self.config, self._include, self._exclude) = load_config(
            serial_port, input_file, **kwargs
        )

        set_pkt_logging(
            pkt_logger,
            file_name=self.config.get(PACKET_LOG),
            cc_stdout=self.config[REDUCE_PROCESSING] >= DONT_CREATE_MESSAGES,
        )

        self.pkt_protocol, self.pkt_transport = None, None
        self.msg_protocol, self.msg_transport = None, None
        self.msg_protocol, self.msg_transport = self.create_client(process_msg)

        self._buffer = deque()
        self._sched_zone = None
        self._sched_lock = Lock()

        # if config.get("ser2net_server"):
        self._relay = None  # ser2net_server relay

        # if self.config[REDUCE_PROCESSING] > 0:
        self.evo = None  # Evohome(controller=config["controller_id"])
        self.systems: List[SystemBase] = []
        self.system_by_id: Dict = {}
        self.devices: List[Device] = []
        self.device_by_id: Dict = {}

        self._schema, self.known_devices = load_schema(self, **kwargs)
        if not self.known_devices:
            self.config[USE_NAMES] = False

        self._prev_msg = None

    def __repr__(self) -> str:
        """Return an unambiguous string representation of this object."""
        return json.dumps(self.schema)

    def __str__(self) -> str:
        """Return a brief readable string representation of this object."""
        return json.dumps(self.schema, indent=2)

    def _setup_event_handlers(self):
        def handle_exception(loop, context):
            """Handle exceptions on any platform."""
            _LOGGER.error("handle_exception(): Caught: %s", context["message"])

            exc = context.get("exception")
            if exc:
                raise exc
            # asyncio.create_task(self.stop())  # TODO: doesn't work here?

        async def handle_sig_posix(sig):
            """Handle signals on posix platform."""
            _LOGGER.debug("Received a signal (%s), processing...", sig.name)

            if sig in (signal.SIGINT, signal.SIGTERM, signal.SIGHUP):
                await self.stop("handle_sig_posix()")  # OK for after tasks.cancel

            elif sig == signal.SIGUSR1:
                _LOGGER.info("Params: \r\n%s", {self.evo.id: self.evo.params})

            elif sig == signal.SIGUSR2:
                _LOGGER.info("Status: \r\n%s", {self.evo.id: self.evo.status})

        def handle_sig_win32(signum, frame):  # can't be async?
            """Handle signals on win32 platform."""
            _LOGGER.debug(
                "Received a signal (%s), processing...", signal.Signals(signum).name
            )

            if signum in (signal.SIGINT, signal.SIGTERM, signal.SIGBREAK):
                # self.stop("handle_sig_win32()")
                raise GracefulExit()

        # signal.SIGBREAK: Int from keyboard (CTRL + BREAK)
        # signal.SIGINT:   Int from keyboard (CTRL + C): to raise KeyboardInterrupt
        # signal.SIGTERM:  Termination signal

        _LOGGER.debug("_setup_event_handlers(): Creating exception handler...")
        self._loop.set_exception_handler(handle_exception)

        _LOGGER.debug("_setup_event_handlers(): Creating signal handlers...")
        signals = [signal.SIGINT, signal.SIGTERM]

        if os.name == "posix":  # full support
            for sig in signals + [signal.SIGHUP, signal.SIGUSR1, signal.SIGUSR2]:
                self._loop.add_signal_handler(
                    sig, lambda sig=sig: asyncio.create_task(handle_sig_posix(sig))
                )
        elif os.name == "nt":  # supported, but YMMV
            _LOGGER.warning("Be aware, YMMV with Windows.")
            for signum in signals + [signal.SIGBREAK]:
                signal.signal(signum, handle_sig_win32)
        else:  # unsupported
            raise RuntimeError("Unsupported OS for this module: %s", os.name)

    async def stop(self, xxx=None) -> None:
        """Perform a graceful shutdown/stop."""

        _LOGGER.warning("stop(): Invoked by: %s, doing housekeeping...", xxx)
        tasks = [t for t in self._tasks if t is not asyncio.current_task()]

        logging.debug(f"stop(): Cancelling {len(tasks)} outstanding async tasks...")
        # [print(t) for t in[asyncio.current_task()]]
        # [print(t) for t in self._tasks]
        # [print(t) for t in tasks]
        # [print(t) for t in asyncio.all_tasks()]
        [task.cancel() for task in tasks]
        await asyncio.gather(*tasks, return_exceptions=False)

        _LOGGER.debug("stop(): Complete.")

    async def start(self) -> None:
        if self.serial_port:  # source of packets is a serial port
            self.pkt_protocol, self.pkt_transport = create_pkt_stack(
                self, self.msg_transport._pkt_receiver, serial_port=self.serial_port
            )
            self._tasks.append(
                self.msg_transport._set_dispatcher(self.pkt_protocol.send_data)
            )

        else:  # if self._input_file:
            self.pkt_protocol, self.pkt_transport = create_pkt_stack(
                self, self.msg_transport._pkt_receiver, packet_log=self._input_file
            )

        if self.pkt_transport.get_extra_info(POLLER_TASK):
            self._tasks.append(self.pkt_transport.get_extra_info(POLLER_TASK))

        await asyncio.gather(*self._tasks)
        await self.stop("start()")

    def _get_device(self, dev_addr, ctl_addr=None, domain_id=None) -> Device:
        """Return a device (will create it if required).

        Can also set a controller/system (will create as required). If a controller is
        provided, can also set the domain_id as one of: zone_idx, FF (controllers), FC
        (heater_relay), HW (DHW sensor, relay), or None (unknown, TBA).
        """

        def create_system(ctl) -> SystemBase:
            assert ctl.id not in self.system_by_id, f"Duplicate system id: {ctl.id}"
            if ctl.id in self.system_by_id:
                raise LookupError(f"Duplicated system id: {ctl.id}")

            system = SYSTEM_CLASSES.get(ctl.type, System)(self, ctl)

            if not self.config[DISABLE_DISCOVERY]:
                system._discover()  # discover_flag=DISCOVER_ALL)

            return system

        def create_device(dev_addr, **kwargs) -> Device:
            if dev_addr.id in self.device_by_id:
                raise LookupError(f"Duplicated device id: {dev_addr.id}")

            device = DEVICE_CLASSES.get(dev_addr.type, Device)(self, dev_addr, **kwargs)

            if not self.config[DISABLE_DISCOVERY]:
                device._discover()  # discover_flag=DISCOVER_ALL)

            return device

        ctl = None if ctl_addr is None else self._get_device(ctl_addr, domain_id="FF")
        if ctl is not None and self.evo is None:
            self.evo = ctl._evo

        if dev_addr.type in ("18", "63", "--"):  # valid addresses, but not devices
            return
        if dev_addr.id == "01:000001":  # valid addresses, but not devices
            return

        if isinstance(dev_addr, Device):
            device = dev_addr
        else:
            device = self.device_by_id.get(dev_addr.id)

        if device is None:
            device = create_device(dev_addr, ctl=ctl, domain_id=domain_id)
            # if isinstance(device, Controller):
            # if device.is_controller:
            # if dev_addr.type in SYSTEM_CLASSES:
            if dev_addr.type in ("01", "23"):
                device._evo = create_system(device)

        else:  # update the existing device with any metadata
            if ctl is not None:
                device._set_ctl(ctl)

            if domain_id in ("F9", "FA", "FC", "FF"):
                device._domain_id = domain_id
            elif domain_id is not None and ctl is not None:
                device._set_zone(ctl._evo._get_zone(domain_id))

        return device

    @property
    def schema(self) -> dict:
        """Return the global schema."""

        schema = {"main_controller": self.evo._ctl.id if self.evo else None}

        if self.evo:
            schema[self.evo._ctl.id] = self.evo.schema
        for evo in self.systems:
            if evo is not self.evo:
                schema[evo._ctl.id] = evo.schema

        orphans = [d.id for d in self.devices if d.controller is None]
        orphans.sort()
        schema[ATTR_ORPHANS] = orphans

        return schema

    @property
    def params(self) -> dict:
        result = {}

        result["devices"] = {
            d.id: d.params for d in sorted(self.devices, key=lambda x: x.id)
        }

        return result

    @property
    def status(self) -> dict:
        result = {}

        result["devices"] = {
            d.id: d.status for d in sorted(self.devices, key=lambda x: x.id)
        }

        return result

    def create_client(self, msg_handler) -> Tuple[Callable, Callable]:
        """Create a client protocol for the RAMSES-II message transport."""
        return create_msg_stack(self, msg_handler)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""Evohome RF - a RAMSES-II protocol decoder & analyser.

Construct a command (packet that is to be sent).
"""

import asyncio
from datetime import datetime as dt, timedelta as td
from functools import total_ordering
import json
import logging
import struct
from types import SimpleNamespace
from typing import Optional
import zlib

from .const import (
    _dev_mode_,
    CODES_SANS_DOMAIN_ID,
    CODE_SCHEMA,
    COMMAND_FORMAT,
    COMMAND_REGEX,
    HGI_DEVICE,
)
from .exceptions import ExpiredCallbackError
from .helpers import dt_now, extract_addrs

DAY_OF_WEEK = "day_of_week"
HEAT_SETPOINT = "heat_setpoint"
SWITCHPOINTS = "switchpoints"
TIME_OF_DAY = "time_of_day"

SCHEDULE = "schedule"
ZONE_IDX = "zone_idx"

TIMER_SHORT_SLEEP = 0.05
TIMER_LONG_TIMEOUT = td(seconds=60)

FIVE_MINS = td(minutes=5)


Priority = SimpleNamespace(LOW=6, DEFAULT=4, HIGH=2, ASAP=0)

DEV_MODE = _dev_mode_

_LOGGER = logging.getLogger(__name__)
if DEV_MODE:
    _LOGGER.setLevel(logging.DEBUG)


def _pkt_header(pkt: str, rx_header=None) -> Optional[str]:
    """Return the QoS header of a packet."""

    verb = pkt[4:6]
    if rx_header:
        verb = "RP" if verb == "RQ" else " I"  # RQ/RP, or W/I
    code = pkt[41:45]
    src, dst, _ = extract_addrs(pkt)
    addr = dst.id if src.type == "18" else src.id
    payload = pkt[50:]

    header = "|".join((verb, addr, code))

    if code in ("0001", "7FFF") and rx_header:
        return

    if code in ("0005", "000C"):  # zone_idx, device_class
        return "|".join((header, payload[:4]))

    if code == "0404":  # zone_schedule: zone_idx, frag_idx
        return "|".join((header, payload[:2] + payload[10:12]))

    if code == "0418":  # fault_log: log_idx
        if payload == CODE_SCHEMA["0418"]["null_rp"]:
            return header
        return "|".join((header, payload[4:6]))

    if code in CODES_SANS_DOMAIN_ID:  # have no domain_id
        return header

    return "|".join((header, payload[:2]))  # assume has a domain_id


@total_ordering
class Command:
    """The command class."""

    def __init__(self, verb, dest_addr, code, payload, **kwargs) -> None:
        """Initialise the class."""

        self.verb = verb
        self.from_addr = kwargs.get("from_addr", HGI_DEVICE.id)
        self.dest_addr = dest_addr if dest_addr is not None else self.from_addr
        self.code = code
        self.payload = payload

        self._is_valid = None
        if not self.is_valid:
            raise ValueError(f"Invalid parameter values for command: {self}")

        self.callback = kwargs.get("callback", {})  # TODO: use voluptuous
        if self.callback:
            self.callback["args"] = self.callback.get("args", [])
            self.callback["kwargs"] = self.callback.get("kwargs", {})

        self.qos = self._qos
        self.qos.update(kwargs.get("qos", {}))
        self._priority = self.qos["priority"]
        self._priority_dtm = dt_now()  # used for __lt__, etc.

        self._rx_header = None
        self._tx_header = None

    def __str__(self) -> str:
        """Return a brief readable string representation of this object."""

        return COMMAND_FORMAT.format(
            self.verb,
            self.from_addr,
            self.dest_addr,
            self.code,
            int(len(self.payload) / 2),
            self.payload,
        )

    @property
    def _qos(self) -> dict:
        """Return the QoS params of this (request) packet."""

        # the defaults for these are in packet.py
        # qos = {"priority": Priority.DEFAULT, "retries": 3, "timeout": td(seconds=0.5)}
        qos = {"priority": Priority.DEFAULT, "retries": 3}

        if self.code in ("0016", "1F09") and self.verb == "RQ":
            qos.update({"priority": Priority.HIGH, "retries": 5})

        elif self.code == "0404" and self.verb in ("RQ", " W"):
            qos.update({"priority": Priority.HIGH, "timeout": td(seconds=0.30)})

        elif self.code == "0418" and self.verb == "RQ":
            qos.update({"priority": Priority.LOW, "retries": 3})

        return qos

    @property
    def tx_header(self) -> str:
        """Return the QoS header of this (request) packet."""
        if self._tx_header is None:
            self._tx_header = _pkt_header(f"... {self}")
        return self._tx_header

    @property
    def rx_header(self) -> Optional[str]:
        """Return the QoS header of a response packet (if any)."""
        if self.tx_header and self._rx_header is None:
            self._rx_header = _pkt_header(f"... {self}", rx_header=True)
        return self._rx_header

    @property
    def is_valid(self) -> Optional[bool]:
        """Return True if a valid command, otherwise return False/None."""

        if self._is_valid is not None:
            return self._is_valid

        if not COMMAND_REGEX.match(str(self)):
            self._is_valid = False
        elif 0 > len(self.payload) > 96:
            self._is_valid = False
        else:
            self._is_valid = True

        if not self._is_valid:
            _LOGGER.debug("Command has an invalid structure: %s", self)

        return self._is_valid

    @staticmethod
    def _is_valid_operand(other) -> bool:
        return hasattr(other, "_priority") and hasattr(other, "_priority_dtm")

    def __eq__(self, other) -> bool:
        if not self._is_valid_operand(other):
            return NotImplemented
        return (self._priority, self._priority_dtm) == (
            other._priority,
            other._priority_dtm,
        )

    def __lt__(self, other) -> bool:
        if not self._is_valid_operand(other):
            return NotImplemented
        return (self._priority, self._priority_dtm) < (
            other._priority,
            other._priority_dtm,
        )


class FaultLog:  # 0418
    """The fault log of a system."""

    def __init__(self, ctl, msg=None, **kwargs) -> None:
        _LOGGER.debug("FaultLog(ctl=%s).__init__()", ctl)

        self.id = ctl.id
        self._ctl = ctl
        # self._evo = ctl._evo
        self._gwy = ctl._gwy

        self._fault_log = None
        self._fault_log_done = None

        self._limit = 11  # TODO: make configurable

    def __repr_(self) -> str:
        return json.dumps(self._fault_log) if self._fault_log_done else None

    def __str_(self) -> str:
        return f"{self._ctl} (fault log)"

    @property
    def fault_log(self) -> Optional[dict]:
        """Return the fault log of a system."""
        if not self._fault_log_done:
            return

        result = {
            x: {k: v for k, v in y.items() if k[:1] != "_"}
            for x, y in self._fault_log.items()
        }

        return {k: [x for x in v.values()] for k, v in result.items()}

    async def get_fault_log(self, force_refresh=None) -> Optional[dict]:
        """Get the fault log of a system."""
        _LOGGER.debug("FaultLog(%s).get_fault_log()", self)

        self._fault_log = {}
        self._fault_log_done = None

        self._rq_log_entry(log_idx=0)  # calls asyncio.create_task()

        time_start = dt.now()
        while not self._fault_log_done:
            await asyncio.sleep(TIMER_SHORT_SLEEP)
            if dt.now() > time_start + TIMER_LONG_TIMEOUT * 2:
                raise ExpiredCallbackError("failed to obtain log entry (long)")

        return self.fault_log

    def _rq_log_entry(self, log_idx=0):
        """Request the next log entry."""
        _LOGGER.debug("FaultLog(%s)._rq_log_entry(%s)", self, log_idx)

        async def rq_callback(msg) -> None:
            _LOGGER.debug("FaultLog(%s)._proc_log_entry(%s)", self.id, msg)

            if not msg:
                self._fault_log_done = True
                # raise ExpiredCallbackError("failed to obtain log entry (short)")
                return

            if not msg.payload:
                # TODO: delete other callbacks rather than waiting for them to expire
                self._fault_log_done = True
                return

            log = dict(msg.payload)
            log_idx = int(log.pop("log_idx"), 16)
            self._fault_log[log_idx] = log

            if log_idx < self._limit:
                self._rq_log_entry(log_idx + 1)
            else:
                self._fault_log_done = True

        # TODO: (make method) register callback for null response (no log_idx left)
        header = "|".join(("RP", self.id, "0418"))
        if header not in self._gwy.msg_transport._callbacks:
            self._gwy.msg_transport._callbacks[header] = {
                "func": rq_callback,
                "daemon": True,
                "args": [],
                "kwargs": {},
            }

        callback = {"func": rq_callback, "timeout": td(seconds=10)}
        cmd = Command("RQ", self._ctl.id, "0418", f"{log_idx:06X}", callback=callback)
        asyncio.create_task(self._gwy.msg_protocol.send_data(cmd))


class Schedule:  # 0404
    """The schedule of a zone."""

    def __init__(self, zone, **kwargs) -> None:
        _LOGGER.debug("Schedule(zone=%s).__init__()", zone)

        self.id = zone.id
        self._zone = zone
        self.idx = zone.idx

        self._ctl = zone._ctl
        self._evo = zone._evo
        self._gwy = zone._gwy

        self._schedule = None
        self._schedule_done = None

        # initialse the fragment array()
        self._num_frags = None
        self._rx_frags = None
        self._tx_frags = None

    def __repr_(self) -> str:
        return json.dumps(self.schedule) if self._schedule_done else None

    def __str_(self) -> str:
        return f"{self._zone} (schedule)"

    @property
    def schedule(self) -> Optional[dict]:
        """Return the schedule of a zone."""
        if not self._schedule_done or None in self._rx_frags:
            return
        if self._schedule:
            return self._schedule

        if self._rx_frags[0]["msg"].payload["frag_total"] == 255:
            return {}

        frags = [v for d in self._rx_frags for k, v in d.items() if k == "fragment"]

        try:
            self._schedule = self._frags_to_sched(frags)
        except zlib.error:
            self._schedule = None
            _LOGGER.exception("Invalid schedule fragments: %s", frags)
            return

        return self._schedule

    async def get_schedule(self, force_refresh=None) -> Optional[dict]:
        """Get the schedule of a zone."""
        _LOGGER.debug(f"Schedule({self.id}).get_schedule()")

        if not await self._obtain_lock():  # TODO: should raise a TimeOut
            return

        if force_refresh:
            self._schedule_done = None

        if not self._schedule_done:
            self._rq_fragment(frag_cnt=0)  # calls asyncio.create_task()

            time_start = dt.now()
            while not self._schedule_done:
                await asyncio.sleep(TIMER_SHORT_SLEEP)
                if dt.now() > time_start + TIMER_LONG_TIMEOUT:
                    self._release_lock()
                    raise ExpiredCallbackError("failed to get schedule")

        self._release_lock()

        return self.schedule

    def _rq_fragment(self, frag_cnt=0) -> None:
        """Request the next missing fragment (index starts at 1, not 0)."""
        _LOGGER.debug("Schedule(%s)._rq_fragment(%s)", self.id, frag_cnt)

        async def rq_callback(msg) -> None:
            if not msg:  # _LOGGER.debug()... TODO: needs fleshing out
                # TODO: remove any callbacks from msg._gwy.msg_transport._callbacks
                _LOGGER.warning(f"Schedule({self.id}): Callback timed out")
                self._schedule_done = True
                return

            _LOGGER.debug(
                f"Schedule({self.id})._proc_fragment(msg), frag_idx=%s, frag_cnt=%s",
                msg.payload.get("frag_index"),
                msg.payload.get("frag_total"),
            )

            if msg.payload["frag_total"] == 255:  # no schedule (i.e. no zone)
                _LOGGER.warning(f"Schedule({self.id}): No schedule")
                # TODO: remove any callbacks from msg._gwy.msg_transport._callbacks
                pass  # self._rx_frags = [None]

            elif msg.payload["frag_total"] != len(self._rx_frags):  # e.g. 1st frag
                self._rx_frags = [None] * msg.payload["frag_total"]

            self._rx_frags[msg.payload["frag_index"] - 1] = {
                "fragment": msg.payload["fragment"],
                "msg": msg,
            }

            # discard any fragments significantly older that this most recent fragment
            for frag in [f for f in self._rx_frags if f is not None]:
                frag = None if frag["msg"].dtm < msg.dtm - FIVE_MINS else frag

            if None in self._rx_frags:  # there are still frags to get
                self._rq_fragment(frag_cnt=msg.payload["frag_total"])
            else:
                self._schedule_done = True

        if frag_cnt == 0:
            self._rx_frags = [None]  # and: frag_idx = 0

        frag_idx = next((i for i, f in enumerate(self._rx_frags) if f is None), -1)

        # 053 RQ --- 30:185469 01:037519 --:------ 0006 001 00
        # 045 RP --- 01:037519 30:185469 --:------ 0006 004 000500E6

        # 059 RQ --- 30:185469 01:037519 --:------ 0404 007 00-23000800 0100
        # 045 RP --- 01:037519 30:185469 --:------ 0404 048 00-23000829 0104 688...
        # 059 RQ --- 30:185469 01:037519 --:------ 0404 007 00-23000800 0204
        # 045 RP --- 01:037519 30:185469 --:------ 0404 048 00-23000829 0204 4AE...
        # 059 RQ --- 30:185469 01:037519 --:------ 0404 007 00-23000800 0304
        # 046 RP --- 01:037519 30:185469 --:------ 0404 048 00-23000829 0304 6BE...

        payload = f"{self.idx}20000800{frag_idx + 1:02d}{frag_cnt:02d}"  # DHW: 23000800
        callback = {"func": rq_callback, "timeout": td(seconds=1)}
        cmd = Command("RQ", self._ctl.id, "0404", payload, callback=callback)
        asyncio.create_task(self._gwy.msg_protocol.send_data(cmd))

    @staticmethod
    def _frags_to_sched(frags: list) -> dict:
        # _LOGGER.debug(f"Sched({self})._frags_to_sched: array is: %s", frags)
        raw_schedule = zlib.decompress(bytearray.fromhex("".join(frags)))

        zone_idx, schedule = None, []
        old_day, switchpoints = 0, []

        for i in range(0, len(raw_schedule), 20):
            zone_idx, day, time, temp, _ = struct.unpack(
                "<xxxxBxxxBxxxHxxHH", raw_schedule[i : i + 20]
            )
            if day > old_day:
                schedule.append({DAY_OF_WEEK: old_day, SWITCHPOINTS: switchpoints})
                old_day, switchpoints = day, []
            switchpoints.append(
                {
                    TIME_OF_DAY: "{0:02d}:{1:02d}".format(*divmod(time, 60)),
                    HEAT_SETPOINT: temp / 100,
                }
            )

        schedule.append({DAY_OF_WEEK: old_day, SWITCHPOINTS: switchpoints})

        return {ZONE_IDX: f"{zone_idx:02X}", SCHEDULE: schedule}

    @staticmethod
    def _sched_to_frags(schedule: dict) -> list:
        # _LOGGER.debug(f"Sched({self})._sched_to_frags: array is: %s", schedule)
        frags = [
            (
                int(schedule[ZONE_IDX], 16),
                int(week_day[DAY_OF_WEEK]),
                int(setpoint[TIME_OF_DAY][:2]) * 60 + int(setpoint[TIME_OF_DAY][3:]),
                int(setpoint[HEAT_SETPOINT] * 100),
            )
            for week_day in schedule[SCHEDULE]
            for setpoint in week_day[SWITCHPOINTS]
        ]
        frags = [struct.pack("<xxxxBxxxBxxxHxxHxx", *s) for s in frags]

        cobj = zlib.compressobj(level=9, wbits=14)
        blob = b"".join([cobj.compress(s) for s in frags]) + cobj.flush()
        blob = blob.hex().upper()

        return [blob[i : i + 82] for i in range(0, len(blob), 82)]

    async def set_schedule(self, schedule) -> None:
        """Set the schedule of a zone."""
        _LOGGER.debug(f"Schedule({self.id}).set_schedule(schedule)")

        if not await self._obtain_lock():  # TODO: should raise a TimeOut
            return

        self._schedule_done = None

        self._tx_frags = self._sched_to_frags(schedule)
        self._tx_fragment(frag_idx=0)

        time_start = dt.now()
        while not self._schedule_done:
            await asyncio.sleep(TIMER_SHORT_SLEEP)
            if dt.now() > time_start + TIMER_LONG_TIMEOUT:
                self._release_lock()
                raise ExpiredCallbackError("failed to set schedule")

        self._release_lock()

    def _tx_fragment(self, frag_idx=0) -> None:
        """Send the next fragment (index starts at 0)."""
        _LOGGER.debug(
            "Schedule(%s)._tx_fragment(%s/%s)", self.id, frag_idx, len(self._tx_frags)
        )

        async def tx_callback(msg) -> None:
            _LOGGER.debug(
                f"Schedule({self.id})._proc_fragment(msg), frag_idx=%s, frag_cnt=%s",
                msg.payload.get("frag_index"),
                msg.payload.get("frag_total"),
            )

            if msg.payload["frag_index"] < msg.payload["frag_total"]:
                self._tx_fragment(frag_idx=msg.payload.get("frag_index"))
            else:
                self._schedule_done = True

        payload = "{0}200008{1:02X}{2:02d}{3:02d}{4:s}".format(
            self.idx,
            int(len(self._tx_frags[frag_idx]) / 2),
            frag_idx + 1,
            len(self._tx_frags),
            self._tx_frags[frag_idx],
        )
        callback = {"func": tx_callback, "timeout": td(seconds=3)}  # 1 sec too low
        cmd = Command(" W", self._ctl.id, "0404", payload, callback=callback)
        asyncio.create_task(self._gwy.msg_protocol.send_data(cmd))

    async def _obtain_lock(self) -> bool:  # Lock to prevent Rx/Tx at same time
        while True:

            self._evo.zone_lock.acquire()
            if self._evo.zone_lock_idx is None:
                self._evo.zone_lock_idx = self.idx
            self._evo.zone_lock.release()

            if self._evo.zone_lock_idx == self.idx:
                break

            await asyncio.sleep(0.1)  # gives the other zone enough time

        return True

    def _release_lock(self) -> None:
        self._evo.zone_lock.acquire()
        self._evo.zone_lock_idx = None
        self._evo.zone_lock.release()
